"""This module focuses on improvements to the Python libtcod API.
"""
from __future__ import absolute_import as _

import sys as _sys

import warnings

from tcod.libtcod import lib, ffi, BKGND_DEFAULT, BKGND_SET


def _unpack_char_p(char_p):
    if char_p == ffi.NULL:
        return ''
    return ffi.string(char_p).decode()


def _int(int_or_str):
    'return an integer where a single character string may be expected'
    if isinstance(int_or_str, str):
        return ord(int_or_str)
    if isinstance(int_or_str, bytes):
        return int_or_str[0]
    return int(int_or_str) # check for __count__


if _sys.version_info[0] == 2: # Python 2
    def _bytes(string):
        if isinstance(string, unicode):
            return string.encode()
        return string

    def _unicode(string):
        if not isinstance(string, unicode):
            return string.decode('latin-1')
        return string

else: # Python 3
    def _bytes(string):
        if isinstance(string, str):
            return string.encode()
        return string

    def _unicode(string):
        if isinstance(string, bytes):
            return string.decode('latin-1')
        return string


def _fmt_bytes(string):
    return _bytes(string).replace(b'%', b'%%')

def _fmt_unicode(string):
    return _unicode(string).replace(u'%', u'%%')


class _PropagateException():
    """ context manager designed to propagate exceptions outside of a cffi
    callback context.  normally cffi suppresses the exception

    when propagate is called this class will hold onto the error until the
    control flow leaves the context, then the error will be raised

    with _PropagateException as propagate:
    # give propagate as onerror parameter for ffi.def_extern
    """

    def __init__(self):
        self.exc_info = None # (exception, exc_value, traceback)

    def propagate(self, *exc_info):
        """ set an exception to be raised once this context exits

        if multiple errors are caught, only keep the first exception raised
        """
        if not self.exc_info:
            self.exc_info = exc_info

    def __enter__(self):
        """ once in context, only the propagate call is needed to use this
        class effectively
        """
        return self.propagate

    def __exit__(self, type, value, traceback):
        """ if we're holding on to an exception, raise it now

        prefers our held exception over any current raising error

        self.exc_info is reset now in case of nested manager shenanigans
        """
        if self.exc_info:
            type, value, traceback = self.exc_info
            self.exc_info = None
        if type:
            # Python 2/3 compatible throw
            exception = type(value)
            exception.__traceback__ = traceback
            raise exception


class _CDataWrapper(object):

    def __init__(self, *args, **kargs):
        self.cdata = self._get_cdata_from_args(*args, **kargs)
        if self.cdata == None:
            self.cdata = ffi.NULL
        super(_CDataWrapper, self).__init__()

    @staticmethod
    def _get_cdata_from_args(*args, **kargs):
        if len(args) == 1 and isinstance(args[0], ffi.CData) and not kargs:
            return args[0]
        else:
            return None


    def __hash__(self):
        return hash(self.cdata)

    def __eq__(self, other):
        try:
            return self.cdata == other.cdata
        except AttributeError:
            return NotImplemented

    def __getattr__(self, attr):
        if 'cdata' in self.__dict__:
            return getattr(self.__dict__['cdata'], attr)
        raise AttributeError(attr)

    def __setattr__(self, attr, value):
        if hasattr(self, 'cdata') and hasattr(self.cdata, attr):
            setattr(self.cdata, attr, value)
        else:
            super(_CDataWrapper, self).__setattr__(attr, value)


def _console(console):
    """Return a cffi console."""
    try:
        return console.console_c
    except AttributeError:
        warnings.warn(
            ("Falsy console parameters are deprecated, "
             "always use a console instance."),
            DeprecationWarning,
            stacklevel=2,
            )
        return ffi.NULL



"""
    This module handles user input.

    To handle user input you will likely want to use the :any:`event.get`
    function or create a subclass of :any:`event.App`.

    - :any:`tdl.event.get` iterates over recent events.
    - :any:`tdl.event.App` passes events to the overridable methods: ``ev_*``
      and ``key_*``.

    But there are other options such as :any:`event.key_wait` and
    :any:`event.is_window_closed`.

    A few event attributes are actually string constants.
    Here's a reference for those:

    - :any:`Event.type`:
      'QUIT', 'KEYDOWN', 'KEYUP', 'MOUSEDOWN', 'MOUSEUP', or 'MOUSEMOTION.'
    - :any:`MouseButtonEvent.button` (found in :any:`MouseDown` and
      :any:`MouseUp` events):
      'LEFT', 'MIDDLE', 'RIGHT', 'SCROLLUP', 'SCROLLDOWN'
    - :any:`KeyEvent.key` (found in :any:`KeyDown` and :any:`KeyUp` events):
      'NONE', 'ESCAPE', 'BACKSPACE', 'TAB', 'ENTER', 'SHIFT', 'CONTROL',
      'ALT', 'PAUSE', 'CAPSLOCK', 'PAGEUP', 'PAGEDOWN', 'END', 'HOME', 'UP',
      'LEFT', 'RIGHT', 'DOWN', 'PRINTSCREEN', 'INSERT', 'DELETE', 'LWIN',
      'RWIN', 'APPS', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
      'KP0', 'KP1', 'KP2', 'KP3', 'KP4', 'KP5', 'KP6', 'KP7', 'KP8', 'KP9',
      'KPADD', 'KPSUB', 'KPDIV', 'KPMUL', 'KPDEC', 'KPENTER', 'F1', 'F2',
      'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12',
      'NUMLOCK', 'SCROLLLOCK', 'SPACE', 'CHAR'
"""

import time as _time

from tcod import ffi as _ffi
from tcod import lib as _lib

import tdl as _tdl
from . import style as _style

_eventQueue = []
_pushedEvents = []
_eventsflushed = False

_mousel = 0
_mousem = 0
_mouser = 0

# this interprets the constants from libtcod and makes a key -> keyname dictionary
def _parseKeyNames(lib):
    """
    returns a dictionary mapping of human readable key names to their keycodes
    this parses constants with the names of K_* and makes code=name pairs
    this is for KeyEvent.key variable and that enables things like:
    if (event.key == 'PAGEUP'):
    """
    _keyNames = {}
    for attr in dir(lib): # from the modules variables
        if attr[:6] == 'TCODK_': # get the K_* constants
            _keyNames[getattr(lib, attr)] = attr[6:] # and make CODE=NAME pairs
    return _keyNames

_keyNames = _parseKeyNames(_lib)

class Event(object):
    """Base Event class.

    You can easily subclass this to make your own events.  Be sure to set
    the class attribute L{Event.type} for it to be passed to a custom
    :any:`App` ev_* method.
    """
    type = None
    """String constant representing the type of event.

    The :any:`App` ev_* methods depend on this attribute.

    Can be: 'QUIT', 'KEYDOWN', 'KEYUP', 'MOUSEDOWN', 'MOUSEUP',
    or 'MOUSEMOTION.'
    """

    def __repr__(self):
        """List an events public attributes when printed.
        """
        attrdict = {}
        for varname in dir(self):
            if '_' == varname[0]:
                continue
            attrdict[varname] = self.__getattribute__(varname)
        return '%s Event %s' % (self.__class__.__name__, repr(attrdict))

class Quit(Event):
    """Fired when the window is closed by the user.
    """
    __slots__ = ()
    type = 'QUIT'

class KeyEvent(Event):
    """Base class for key events."""

    def __init__(self, key='', char='', text='', shift=False,
                 left_alt=False, right_alt=False,
                 left_control=False, right_control=False,
                 left_meta=False, right_meta=False):
        # Convert keycodes into string, but use string if passed
        self.key = key if isinstance(key, str) else _keyNames[key]
        """Text: Human readable names of the key pressed.
        Non special characters will show up as 'CHAR'.

        Can be one of
        'NONE', 'ESCAPE', 'BACKSPACE', 'TAB', 'ENTER', 'SHIFT', 'CONTROL',
        'ALT', 'PAUSE', 'CAPSLOCK', 'PAGEUP', 'PAGEDOWN', 'END', 'HOME', 'UP',
        'LEFT', 'RIGHT', 'DOWN', 'PRINTSCREEN', 'INSERT', 'DELETE', 'LWIN',
        'RWIN', 'APPS', '0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
        'KP0', 'KP1', 'KP2', 'KP3', 'KP4', 'KP5', 'KP6', 'KP7', 'KP8', 'KP9',
        'KPADD', 'KPSUB', 'KPDIV', 'KPMUL', 'KPDEC', 'KPENTER', 'F1', 'F2',
        'F3', 'F4', 'F5', 'F6', 'F7', 'F8', 'F9', 'F10', 'F11', 'F12',
        'NUMLOCK', 'SCROLLLOCK', 'SPACE', 'CHAR'

        For the actual character instead of 'CHAR' use :any:`keychar`.
        """
        self.char = char.replace('\x00', '') # change null to empty string
        """Text: A single character string of the letter or symbol pressed.

        Special characters like delete and return are not cross-platform.
        L{key} or L{keychar} should be used instead for special keys.
        Characters are also case sensitive.
        """
        # get the best out of self.key and self.char
        self.keychar = self.char if self.key == 'CHAR' else self.key
        """Similar to L{key} but returns a case sensitive letter or symbol
        instead of 'CHAR'.

        This variable makes available the widest variety of symbols and should
        be used for key-mappings or anywhere where a narrower sample of keys
        isn't needed.
        """
        self.text = text

        self.left_alt = self.leftAlt = bool(left_alt)
        """bool:"""
        self.right_alt = self.rightAlt = bool(right_alt)
        """bool:"""
        self.left_control = self.leftCtrl = bool(left_control)
        """bool:"""
        self.right_control = self.rightCtrl = bool(right_control)
        """bool:"""
        self.shift = bool(shift)
        """bool: True if shift was held down during this event."""
        self.alt = self.left_alt or self.right_alt
        """bool: True if alt was held down during this event."""
        self.control = self.left_control or self.right_control
        """bool: True if control was held down during this event."""
        self.left_meta = bool(left_meta)
        self.right_meta = bool(right_meta)
        self.meta = self.left_meta or self.right_meta

    def __repr__(self):
        parameters = []
        for attr in ('key', 'char', 'text', 'shift',
                     'left_alt', 'right_alt',
                     'left_control', 'right_control',
                     'left_meta', 'right_meta'):
            value = getattr(self, attr)
            if value:
                parameters.append('%s=%r' % (attr, value))
        return '%s(%s)' % (self.__class__.__name__, ', '.join(parameters))

class KeyDown(KeyEvent):
    """Fired when the user presses a key on the keyboard or a key repeats.
    """
    type = 'KEYDOWN'

class KeyUp(KeyEvent):
    """Fired when the user releases a key on the keyboard.
    """
    type = 'KEYUP'

_mouseNames = {1: 'LEFT', 2: 'MIDDLE', 3: 'RIGHT', 4: 'SCROLLUP', 5: 'SCROLLDOWN'}
class MouseButtonEvent(Event):
    """Base class for mouse button events."""

    def __init__(self, button, pos, cell):
        self.button = _mouseNames[button]
        """Text: Can be one of
        'LEFT', 'MIDDLE', 'RIGHT', 'SCROLLUP', 'SCROLLDOWN'
        """
        self.pos = pos
        """Tuple[int, int]: (x, y) position of the mouse on the screen."""
        self.cell = cell
        """Tuple[int, int]: (x, y) position of the mouse snapped to a cell on
        the root console
        """

class MouseDown(MouseButtonEvent):
    """Fired when a mouse button is pressed."""
    __slots__ = ()
    type = 'MOUSEDOWN'

class MouseUp(MouseButtonEvent):
    """Fired when a mouse button is released."""
    __slots__ = ()
    type = 'MOUSEUP'

class MouseMotion(Event):
    """Fired when the mouse is moved."""
    type = 'MOUSEMOTION'

    def __init__(self, pos, cell, motion, cellmotion):
        self.pos = pos
        """(x, y) position of the mouse on the screen.
        type: (int, int)"""
        self.cell = cell
        """(x, y) position of the mouse snapped to a cell on the root console.
        type: (int, int)"""
        self.motion = motion
        """(x, y) motion of the mouse on the screen.
        type: (int, int)"""
        self.cellmotion = cellmotion
        """(x, y) mostion of the mouse moving over cells on the root console.
        type: (int, int)"""

class App(object):
    """
    Application framework.

     - ev_*: Events are passed to methods based on their :any:`Event.type`
       attribute.
       If an event type is 'KEYDOWN' the ev_KEYDOWN method will be called
       with the event instance as a parameter.

     - key_*: When a key is pressed another method will be called based on the
       :any:`KeyEvent.key` attribute.  For example the 'ENTER' key will call
       key_ENTER with the associated :any:`KeyDown` event as its parameter.

     - :any:`update`: This method is called every loop.  It is passed a single
       parameter detailing the time in seconds since the last update
       (often known as deltaTime.)

       You may want to call drawing routines in this method followed by
       :any:`tdl.flush`.

    """
    __slots__ = ('__running', '__prevTime')

    def ev_QUIT(self, event):
        """Unless overridden this method raises a SystemExit exception closing
        the program."""
        raise SystemExit()

    def ev_KEYDOWN(self, event):
        """Override this method to handle a :any:`KeyDown` event."""

    def ev_KEYUP(self, event):
        """Override this method to handle a :any:`KeyUp` event."""

    def ev_MOUSEDOWN(self, event):
        """Override this method to handle a :any:`MouseDown` event."""

    def ev_MOUSEUP(self, event):
        """Override this method to handle a :any:`MouseUp` event."""

    def ev_MOUSEMOTION(self, event):
        """Override this method to handle a :any:`MouseMotion` event."""

    def update(self, deltaTime):
        """Override this method to handle per frame logic and drawing.

        Args:
            deltaTime (float):
                This parameter tells the amount of time passed since
                the last call measured in seconds as a floating point
                number.

                You can use this variable to make your program
                frame rate independent.
                Use this parameter to adjust the speed of motion,
                timers, and other game logic.
        """
        pass

    def suspend(self):
        """When called the App will begin to return control to where
        :any:`App.run` was called.

        Some further events are processed and the :any:`App.update` method
        will be called one last time before exiting
        (unless suspended during a call to :any:`App.update`.)
        """
        self.__running = False

    def run(self):
        """Delegate control over to this App instance.  This function will
        process all events and send them to the special methods ev_* and key_*.

        A call to :any:`App.suspend` will return the control flow back to where
        this function is called.  And then the App can be run again.
        But a single App instance can not be run multiple times simultaneously.
        """
        if getattr(self, '_App__running', False):
            raise _tdl.TDLError('An App can not be run multiple times simultaneously')
        self.__running = True
        while self.__running:
            self.runOnce()

    def run_once(self):
        """Pump events to this App instance and then return.

        This works in the way described in :any:`App.run` except it immediately
        returns after the first :any:`update` call.

        Having multiple :any:`App` instances and selectively calling runOnce on
        them is a decent way to create a state machine.
        """
        if not hasattr(self, '_App__prevTime'):
            self.__prevTime = _time.clock() # initiate __prevTime
        for event in get():
            if event.type: # exclude custom events with a blank type variable
                # call the ev_* methods
                method = 'ev_%s' % event.type # ev_TYPE
                getattr(self, method)(event)
            if event.type == 'KEYDOWN':
                # call the key_* methods
                method = 'key_%s' % event.key # key_KEYNAME
                if hasattr(self, method): # silently exclude undefined methods
                    getattr(self, method)(event)
        newTime = _time.clock()
        self.update(newTime - self.__prevTime)
        self.__prevTime = newTime
        #_tdl.flush()

def _processEvents():
    """Flushes the event queue from libtcod into the global list _eventQueue"""
    global _mousel, _mousem, _mouser, _eventsflushed, _pushedEvents
    _eventsflushed = True
    events = _pushedEvents # get events from event.push
    _pushedEvents = [] # then clear the pushed events queue

    mouse = _ffi.new('TCOD_mouse_t *')
    libkey = _ffi.new('TCOD_key_t *')
    while 1:
        libevent = _lib.TCOD_sys_check_for_event(_lib.TCOD_EVENT_ANY, libkey, mouse)
        if not libevent: # no more events from libtcod
            break

        #if mouse.dx or mouse.dy:
        if libevent & _lib.TCOD_EVENT_MOUSE_MOVE:
            events.append(MouseMotion((mouse.x, mouse.y),
                                      (mouse.cx, mouse.cy),
                                      (mouse.dx, mouse.dy),
                                      (mouse.dcx, mouse.dcy)))

        mousepos = ((mouse.x, mouse.y), (mouse.cx, mouse.cy))

        for oldstate, newstate, released, button in \
            zip((_mousel, _mousem, _mouser),
                (mouse.lbutton, mouse.mbutton, mouse.rbutton),
                (mouse.lbutton_pressed, mouse.mbutton_pressed,
                 mouse.rbutton_pressed),
                (1, 2, 3)):
            if released:
                if not oldstate:
                    events.append(MouseDown(button, *mousepos))
                events.append(MouseUp(button, *mousepos))
                if newstate:
                    events.append(MouseDown(button, *mousepos))
            elif newstate and not oldstate:
                events.append(MouseDown(button, *mousepos))

        if mouse.wheel_up:
            events.append(MouseDown(4, *mousepos))
        if mouse.wheel_down:
            events.append(MouseDown(5, *mousepos))

        _mousel = mouse.lbutton
        _mousem = mouse.mbutton
        _mouser = mouse.rbutton

        if libkey.vk == _lib.TCODK_NONE:
            break
        if libkey.pressed:
            keyevent = KeyDown
        else:
            keyevent = KeyUp

        events.append(
            keyevent(
                libkey.vk,
                libkey.c.decode('ascii', errors='ignore'),
                _ffi.string(libkey.text).decode('utf-8'),
                libkey.shift,
                libkey.lalt,
                libkey.ralt,
                libkey.lctrl,
                libkey.rctrl,
                libkey.lmeta,
                libkey.rmeta,
                )
            )

    if _lib.TCOD_console_is_window_closed():
        events.append(Quit())

    _eventQueue.extend(events)

def get():
    """Flushes the event queue and returns the list of events.

    This function returns :any:`Event` objects that can be identified by their
    type attribute or their class.

    Returns: Iterator[Type[Event]]: An iterable of Events or anything
        put in a :any:`push` call.

        If the iterator is deleted or otherwise interrupted before finishing
        the excess items are preserved for the next call.
    """
    _processEvents()
    return _event_generator()

def _event_generator():
    while _eventQueue:
        # if there is an interruption the rest of the events stay untouched
        # this means you can break out of a event.get loop without losing
        # the leftover events
        yield(_eventQueue.pop(0))
    raise StopIteration()


def wait(timeout=None, flush=True):
    """Wait for an event.

    Args:
        timeout (Optional[int]): The time in seconds that this function will
            wait before giving up and returning None.

            With the default value of None, this will block forever.
        flush (bool): If True a call to :any:`tdl.flush` will be made before
            listening for events.

    Returns: Type[Event]: An event, or None if the function
             has timed out.
             Anything added via :any:`push` will also be returned.
    """
    if timeout is not None:
        timeout = timeout + _time.clock() # timeout at this time
    while True:
        if _eventQueue:
            return _eventQueue.pop(0)
        if flush:
            # a full 'round' of events need to be processed before flushing
            _tdl.flush()
        if timeout and _time.clock() >= timeout:
            return None # return None on timeout
        _time.sleep(0.001) # sleep 1ms
        _processEvents()


def push(event):
    """Push an event into the event buffer.

    Args:
        event (Any): This event will be available on the next call to
            :any:`event.get`.

            An event pushed in the middle of a :any:`get` will not show until
            the next time :any:`get` called preventing push related
            infinite loops.

            This object should at least have a 'type' attribute.
    """
    _pushedEvents.append(event)

def key_wait():
    """Waits until the user presses a key.
    Then returns a :any:`KeyDown` event.

    Key events will repeat if held down.

    A click to close the window will be converted into an Alt+F4 KeyDown event.

    Returns:
        tdl.event.KeyDown: The pressed key.
    """
    while 1:
        for event in get():
            if event.type == 'KEYDOWN':
                return event
            if event.type == 'QUIT':
                # convert QUIT into alt+F4
                return KeyDown('F4', '', True, False, True, False, False)
        _time.sleep(.001)

def set_key_repeat(delay=500, interval=0):
    """Does nothing.
    """
    pass

def is_window_closed():
    """Returns True if the exit button on the window has been clicked and
    stays True afterwards.

    Returns: bool:
    """
    return _lib.TCOD_console_is_window_closed()

__all__ = [_var for _var in locals().keys() if _var[0] != '_']

App.runOnce = _style.backport(App.run_once)
keyWait = _style.backport(key_wait)
setKeyRepeat = _style.backport(set_key_repeat)
isWindowClosed = _style.backport(is_window_closed)




"""This module handles loading of the libtcod cffi API.
"""
from __future__ import absolute_import as _

import sys as _sys
import os as _os

import platform as _platform


from tcod import __path__

if _sys.platform == 'win32':
    # add Windows dll's to PATH
    _bits, _linkage = _platform.architecture()
    _os.environ['PATH'] = '%s;%s' % (
        _os.path.join(__path__[0], 'x86' if _bits == '32bit' else 'x64'),
        _os.environ['PATH'],
        )

def _import_library_functions(lib):
    # imports libtcod namespace into thie module
    # does not override existing names
    g = globals()
    for name in dir(lib):
        if name[:5] == 'TCOD_':
            if name.isupper():
                g[name[5:]] = getattr(lib, name) # const names
        elif name.startswith('FOV'):
            g[name] = getattr(lib, name) # fov const names
        elif name[:6] == 'TCODK_': # key name
            g['KEY_' + name[6:]] = getattr(lib, name)

NOISE_DEFAULT_HURST = 0.5
NOISE_DEFAULT_LACUNARITY = 2.0

def FOV_PERMISSIVE(p) :
    return FOV_PERMISSIVE_0+p

def BKGND_ALPHA(a):
    return BKGND_ALPH | (int(a * 255) << 8)

def BKGND_ADDALPHA(a):
    return BKGND_ADDA | (int(a * 255) << 8)

class _Mock(object):
    """Mock object needed for ReadTheDocs."""

    TCOD_RENDERER_GLSL = 0
    TCOD_RENDERER_OPENGL = 1
    TCOD_RENDERER_SDL = 2
    TCOD_FONT_LAYOUT_ASCII_INCOL = 1
    TCOD_BKGND_SET = 1
    TCOD_BKGND_DEFAULT = 13
    TCOD_KEY_RELEASED = 2
    TCOD_NOISE_DEFAULT = 0
    TCOD_NOISE_SIMPLEX = 2
    TCOD_NOISE_WAVELET = 4
    FOV_RESTRICTIVE = 12

    TCOD_RNG_MT = 0
    TCOD_RNG_CMWC = 1

    CData = () # This gets passed to an isinstance call.

    @staticmethod
    def def_extern():
        """Pass def_extern call silently."""
        return lambda func:func

    def __getattr__(self, attr):
        """This object pretends to have everything."""
        return self

    def __call__(self, *args, **kargs):
        """Suppress any other calls"""
        return self

    def __str__(self):
        """Just have ? in case anything leaks as a parameter default."""
        return '?'


if _os.environ.get('READTHEDOCS'):
    # Mock the lib and ffi objects needed to compile docs for readthedocs.io
    # Allows an import without building the cffi module first.
    lib = ffi = _Mock()
else:
    from tcod._libtcod import lib, ffi

_import_library_functions(lib)



"""
    This module provides a simple CFFI API to libtcod.

    This port has large partial support for libtcod's C functions.
    Use tcod/libtcod_cdef.h in the source distribution to see specially what
    functions were exported and what new functions have been added by TDL.

    The ffi and lib variables should be familiar to anyone that has used CFFI
    before, otherwise it's time to read up on how they work:
    https://cffi.readthedocs.org/en/latest/using.html

    Otherwise this module can be used as a drop in replacement for the official
    libtcod.py module.

    Bring any issues or requests to GitHub:
    https://github.com/HexDecimal/libtcod-cffi
"""
from __future__ import absolute_import

from tcod.libtcodpy import *
from tcod.version import __version__



"""
    Random module docs.
"""

from __future__ import absolute_import as _

import time

from tcod.libtcod import ffi, lib
from tcod.libtcod import RNG_MT as MERSENNE_TWISTER
from tcod.libtcod import RNG_CMWC as COMPLEMENTARY_MULTIPLY_WITH_CARRY


class Random(object):
    """The libtcod random number generator.

    If all you need is a random number generator then it's recommended
    that you use the :any:`random` module from the Python standard library.

    If ``seed`` is None then a random seed will be generated.

    Args:
        algorithm (int): The algorithm to use.
        seed (Optional[Hashable]):
            Could be a 32-bit integer, but any hashable object is accepted.

    Attributes:
        random_c (CData): A cffi pointer to a TCOD_random_t object.
    """
    def __init__(self, algorithm, seed=None):
        """Create a new instance using this algorithm and seed."""
        if seed is None:
            seed = time.time() + time.clock()
        self.random_c = ffi.gc(
            ffi.cast('mersenne_data_t*',
                     lib.TCOD_random_new_from_seed(algorithm,
                                                   hash(seed) % (1 << 32))),
            lib.TCOD_random_delete)

    @classmethod
    def _new_from_cdata(cls, cdata):
        """Return a new instance encapsulating this cdata."""
        self = object.__new__(cls)
        self.random_c = cdata
        return self

    def randint(self, low, high):
        """Return a random integer within the linear range: low <= n <= high.

        Args:
            low (int): The lower bound of the random range.
            high (int): The upper bound of the random range.

        Returns:
            int: A random integer.
        """
        return lib.TCOD_random_get_i(self.random_c, low, high)

    def uniform(self, low, high):
        """Return a random floating number in the range: low <= n <= high.

        Args:
            low (int): The lower bound of the random range.
            high (int): The upper bound of the random range.

        Returns:
            float: A random float.
        """
        return lib.TCOD_random_get_double(self.random_c, low, high)

    def guass(self, mu, sigma):
        """Return a random number using Gaussian distribution.

        Args:
            mu (float): The median returned value.
            sigma (float): The standard deviation.

        Returns:
            float: A random float.
        """
        return lib.TCOD_random_get_gaussian_double(self.random_c, mu, sigma)

    def inverse_guass(self, mu, sigma):
        """Return a random Gaussian number using the Box-Muller transform.

        Args:
            mu (float): The median returned value.
            sigma (float): The standard deviation.

        Returns:
            float: A random float.
        """
        return lib.TCOD_random_get_gaussian_double_inv(self.random_c, mu, sigma)

    def __getstate__(self):
        """Pack the self.random_c attribute into a portable state."""
        state = self.__dict__.copy()
        state['random_c'] = {
            'algo': self.random_c.algo,
            'distribution': self.random_c.distribution,
            'mt': list(self.random_c.mt),
            'cur_mt': self.random_c.cur_mt,
            'Q': list(self.random_c.Q),
            'c': self.random_c.c,
            'cur': self.random_c.cur,
            }
        return state

    def __setstate__(self, state):
        """Create a new cdata object with the stored paramaters."""
        try:
            cdata = state['random_c']
        except KeyError: # old/deprecated format
            cdata = state['cdata']
            del state['cdata']
        state['random_c'] = ffi.new('mersenne_data_t*', cdata)
        self.__dict__.update(state)



"""
libtcod works with a special 'root' console.  You create this console using
the :any:`tcod.console_init_root` function.  Usually after setting the font
with :any:`console_set_custom_font` first.

Example::

    # Make sure 'arial10x10.png' is in the same directory as this script.
    import time

    import tcod

    # Setup the font.
    tcod.console_set_custom_font(
        'arial10x10.png',
        tcod.FONT_LAYOUT_ASCII_INROW | tcod.FONT_LAYOUT_TCOD,
        )
    # Initialize the root console in a context.
    with tcod.console_init_root(80, 60, 'title') as root_console:
        root_console.print_(x=0, y=0, string='Hello World!')
        tcod.console_flush() # Show the console.
        time.sleep(3) # Wait 3 seconds.
    # The window is closed here, after the above context exits.
"""

from __future__ import absolute_import

import sys

import warnings

import numpy as np

import tcod.libtcod
from tcod.libtcod import ffi, lib
import tcod._internal

if sys.version_info[0] == 2: # Python 2
    def _fmt(string):
        if not isinstance(string, unicode):
            string = string.decode('latin-1')
        return string.replace(u'%', u'%%')
else:
    def _fmt(string):
        """Return a string that escapes 'C printf' side effects."""
        return string.replace('%', '%%')


class Console(object):
    """A console object containing a grid of characters with
    foreground/background colors.

    .. versionchanged:: 4.3
        Added `order` parameter.

    Args:
        width (int): Width of the new Console.
        height (int): Height of the new Console.
        order (str): Which numpy memory order to use.

    Attributes:
        console_c (CData): A cffi pointer to a TCOD_console_t object.
    """

    def __init__(self, width, height, order='C'):
        self._key_color = None
        self._ch = np.zeros((height, width), dtype=np.intc)
        self._fg = np.zeros((height, width), dtype='(3,)u1')
        self._bg = np.zeros((height, width), dtype='(3,)u1')
        self._order = tcod._internal.verify_order(order)

        # libtcod uses the root console for defaults.
        bkgnd_flag = alignment = 0
        if lib.TCOD_ctx.root != ffi.NULL:
            bkgnd_flag = lib.TCOD_ctx.root.bkgnd_flag
            alignment = lib.TCOD_ctx.root.alignment

        self._console_data = self.console_c = ffi.new(
            'TCOD_console_data_t*',
            {
            'w': width, 'h': height,
            'ch_array': ffi.cast('int*', self._ch.ctypes.data),
            'fg_array': ffi.cast('TCOD_color_t*', self._fg.ctypes.data),
            'bg_array': ffi.cast('TCOD_color_t*', self._bg.ctypes.data),
            'bkgnd_flag': bkgnd_flag,
            'alignment': alignment,
            'fore': (255, 255, 255),
            'back': (0, 0, 0),
            },
        )

    @classmethod
    def _from_cdata(cls, cdata, order='C'):
        if isinstance(cdata, cls):
            return cdata
        self = object.__new__(cls)
        self.console_c = cdata
        self._init_setup_console_data(order)
        return self

    def _init_setup_console_data(self, order='C'):
        """Setup numpy arrays over libtcod data buffers."""
        self._key_color = None
        if self.console_c == ffi.NULL:
            self._console_data = lib.TCOD_ctx.root
        else:
            self._console_data = ffi.cast('TCOD_console_data_t *', self.console_c)

        def unpack_color(color_data):
            """return a (height, width, 3) shaped array from an image struct"""
            color_buffer = ffi.buffer(color_data[0:self.width * self.height])
            array = np.frombuffer(color_buffer, np.uint8)
            return array.reshape((self.height, self.width, 3))

        self._fg = unpack_color(self._console_data.fg_array)
        self._bg = unpack_color(self._console_data.bg_array)

        buf = self._console_data.ch_array
        buf = ffi.buffer(buf[0:self.width * self.height])
        self._ch = np.frombuffer(buf, np.intc).reshape((self.height,
                                                        self.width))

        self._order = tcod._internal.verify_order(order)

    @property
    def width(self):
        """int: The width of this Console. (read-only)"""
        return lib.TCOD_console_get_width(self.console_c)

    @property
    def height(self):
        """int: The height of this Console. (read-only)"""
        return lib.TCOD_console_get_height(self.console_c)

    @property
    def bg(self):
        """A uint8 array with the shape (height, width, 3).

        You can change the consoles background colors by using this array.

        Index this array with ``console.bg[y, x, channel]``
        """
        return self._bg.transpose(1, 0, 2) if self._order == 'F' else self._bg

    @property
    def fg(self):
        """A uint8 array with the shape (height, width, 3).

        You can change the consoles foreground colors by using this array.

        Index this array with ``console.fg[y, x, channel]``
        """
        return self._fg.transpose(1, 0, 2) if self._order == 'F' else self._fg

    @property
    def ch(self):
        """An integer array with the shape (height, width).

        You can change the consoles character codes by using this array.

        Index this array with ``console.ch[y, x]``
        """
        return self._ch.T if self._order == 'F' else self._ch

    @property
    def default_bg(self):
        """Tuple[int, int, int]: The default background color."""
        color = self._console_data.back
        return color.r, color.g, color.b
    @default_bg.setter
    def default_bg(self, color):
        self._console_data.back = color

    @property
    def default_fg(self):
        """Tuple[int, int, int]: The default foreground color."""
        color = self._console_data.fore
        return color.r, color.g, color.b
    @default_fg.setter
    def default_fg(self, color):
        self._console_data.fore = color

    @property
    def default_bg_blend(self):
        """int: The default blending mode."""
        return self._console_data.bkgnd_flag
    @default_bg_blend.setter
    def default_bg_blend(self, value):
        self._console_data.bkgnd_flag = value

    @property
    def default_alignment(self):
        """int: The default text alignment."""
        return self._console_data.alignment
    @default_alignment.setter
    def default_alignment(self, value):
        self._console_data.alignment = value

    def clear(self):
        """Reset this console to its default colors and the space character.
        """
        lib.TCOD_console_clear(self.console_c)

    def put_char(self, x, y, ch, bg_blend=tcod.libtcod.BKGND_DEFAULT):
        """Draw the character c at x,y using the default colors and a blend mode.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            ch (int): Character code to draw.  Must be in integer form.
            bg_blend (int): Blending mode to use, defaults to BKGND_DEFAULT.
        """
        lib.TCOD_console_put_char(self.console_c, x, y, ch, bg_blend)

    def print_(self, x, y, string, bg_blend=tcod.libtcod.BKGND_DEFAULT,
               alignment=None):
        """Print a color formatted string on a console.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            string (Text): A Unicode string optionaly using color codes.
            bg_blend (int): Blending mode to use, defaults to BKGND_DEFAULT.
            alignment (Optinal[int]): Text alignment.
        """
        alignment = self.default_alignment if alignment is None else alignment

        lib.TCOD_console_print_ex_utf(self.console_c, x, y,
                                      bg_blend, alignment, _fmt(string))

    def print_rect(self, x, y, width, height, string,
                   bg_blend=tcod.libtcod.BKGND_DEFAULT, alignment=None):
        """Print a string constrained to a rectangle.

        If h > 0 and the bottom of the rectangle is reached,
        the string is truncated. If h = 0,
        the string is only truncated if it reaches the bottom of the console.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            width (int): Maximum width to render the text.
            height (int): Maximum lines to render the text.
            string (Text): A Unicode string.
            bg_blend (int): Background blending flag.
            alignment (Optional[int]): Alignment flag.

        Returns:
            int: The number of lines of text once word-wrapped.
        """
        alignment = self.default_alignment if alignment is None else alignment
        return lib.TCOD_console_print_rect_ex_utf(self.console_c,
            x, y, width, height, bg_blend, alignment, _fmt(string))

    def get_height_rect(self, x, y, width, height, string):
        """Return the height of this text word-wrapped into this rectangle.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            width (int): Maximum width to render the text.
            height (int): Maximum lines to render the text.
            string (Text): A Unicode string.

        Returns:
            int: The number of lines of text once word-wrapped.
        """
        return lib.TCOD_console_get_height_rect_utf(
            self.console_c, x, y, width, height, _fmt(string))

    def rect(self, x, y, width, height, clear,
             bg_blend=tcod.libtcod.BKGND_DEFAULT):
        """Draw a the background color on a rect optionally clearing the text.

        If clr is True the affected tiles are changed to space character.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            width (int): Maximum width to render the text.
            height (int): Maximum lines to render the text.
            clear (bool): If True all text in the affected area will be
                          removed.
            bg_blend (int): Background blending flag.
        """
        lib.TCOD_console_rect(self.console_c, x, y, width, height, clear,
                              bg_blend)

    def hline(self, x, y, width, bg_blend=tcod.libtcod.BKGND_DEFAULT):
        """Draw a horizontal line on the console.

        This always uses the character 196, the horizontal line character.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            width (int): The horozontal length of this line.
            bg_blend (int): The background blending flag.
        """
        lib.TCOD_console_hline(self.console_c, x, y, width, bg_blend)

    def vline(self, x, y, height, bg_blend=tcod.libtcod.BKGND_DEFAULT):
        """Draw a vertical line on the console.

        This always uses the character 179, the vertical line character.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            height (int): The horozontal length of this line.
            bg_blend (int): The background blending flag.
        """
        lib.TCOD_console_vline(self.console_c, x, y, height, bg_blend)

    def print_frame(self, x, y, width, height, string='',
                    clear=True, bg_blend=tcod.libtcod.BKGND_DEFAULT):
        """Draw a framed rectangle with optinal text.

        This uses the default background color and blend mode to fill the
        rectangle and the default foreground to draw the outline.

        string will be printed on the inside of the rectangle, word-wrapped.

        Args:
            x (int): The x coordinate from the left.
            y (int): The y coordinate from the top.
            width (int): The width if the frame.
            height (int): The height of the frame.
            string (Text): A Unicode string to print.
            clear (bool): If True all text in the affected area will be
                          removed.
            bg_blend (int): The background blending flag.

        Note:
            This method does not support Unicode outside of the 0-255 range.
        """
        lib.TCOD_console_print_frame(self.console_c, x, y, width, height,
                                     clear, bg_blend, string.encode('latin-1'))

    def blit(self, dest, dest_x=0, dest_y=0,
             src_x=0, src_y=0, width=0, height=0,
             fg_alpha=1.0, bg_alpha=1.0, key_color=None):
        """Blit from this console onto the ``dest`` console.

        Args:
            dest (Console): The destintaion console to blit onto.
            dest_x (int): Leftmost coordinate of the destintaion console.
            dest_y (int): Topmost coordinate of the destintaion console.
            src_x (int): X coordinate from this console to blit, from the left.
            src_y (int): Y coordinate from this console to blit, from the top.
            width (int): The width of the region to blit.

                If this is 0 the maximum possible width will be used.
            height (int): The height of the region to blit.

                If this is 0 the maximum possible height will be used.
            fg_alpha (float): Foreground color alpha vaule.
            bg_alpha (float): Background color alpha vaule.
            key_color (Optional[Tuple[int, int, int]]):
                None, or a (red, green, blue) tuple with values of 0-255.

        .. versionchanged:: 4.0
            Parameters were rearraged and made optional.

            Previously they were:
            `(x, y, width, height, dest, dest_x, dest_y, *)`
        """
        # The old syntax is easy to detect and correct.
        if hasattr(src_y, 'console_c'):
            src_x, src_y, width, height, dest, dest_x, dest_y = \
                dest, dest_x, dest_y, src_x, src_y, width, height
            warnings.warn(
                "Parameter names have been moved around, see documentation.",
                DeprecationWarning,
                )

        if key_color or self._key_color:
            key_color = ffi.new('TCOD_color_t*', key_color)
            lib.TCOD_console_blit_key_color(
            self.console_c, src_x, src_y, width, height,
            dest.console_c, dest_x, dest_y, fg_alpha, bg_alpha, key_color
            )
        else:
            lib.TCOD_console_blit(
                self.console_c, src_x, src_y, width, height,
                dest.console_c, dest_x, dest_y, fg_alpha, bg_alpha
            )

    def set_key_color(self, color):
        """Set a consoles blit transparent color.

        Args:
            color (Tuple[int, int, int]):
        """
        self._key_color = color

    def __enter__(self):
        """Returns this console in a managed context.

        When the root console is used as a context, the graphical window will
        close once the context is left as if :any:`tcod.console_delete` was
        called on it.

        This is useful for some Python IDE's like IDLE, where the window would
        not be closed on its own otherwise.
        """
        if self.console_c != ffi.NULL:
            raise NotImplementedError('Only the root console has a context.')
        return self

    def __exit__(self, *args):
        """Closes the graphical window on exit.

        Some tcod functions may have undefined behavior after this point.
        """
        lib.TCOD_console_delete(self.console_c)

    def __bool__(self):
        """Returns False if this is the root console.

        This mimics libtcodpy behavior.
        """
        return self.console_c != ffi.NULL

    __nonzero__ = __bool__

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['console_c']
        state['_console_data'] = {
            'w': self.width, 'h': self.height,
            'bkgnd_flag': self.default_bg_blend,
            'alignment': self.default_alignment,
            'fore': self.default_fg,
            'back': self.default_bg,
        }
        if self.console_c == ffi.NULL:
            state['_ch'] = np.copy(self._ch)
            state['_fg'] = np.copy(self._fg)
            state['_bg'] = np.copy(self._bg)
        return state

    def __setstate__(self, state):
        self._key_color = None
        self.__dict__.update(state)
        self._console_data.update(
            {
            'ch_array': ffi.cast('int*', self._ch.ctypes.data),
            'fg_array': ffi.cast('TCOD_color_t*', self._fg.ctypes.data),
            'bg_array': ffi.cast('TCOD_color_t*', self._bg.ctypes.data),
            }
        )
        self._console_data = self.console_c = ffi.new(
            'TCOD_console_data_t*', self._console_data)



"""
    Used internally to handle style changes without breaking backwards
    compatibility.
"""

import warnings as _warnings
import functools as _functools

def backport(func):
    """
        Backport a function name into an old style for compatibility.

        The docstring is updated to reflect that the new function returned is
        deprecated and that the other function is preferred.
        A DeprecationWarning is also raised for using this function.

        If the script is run with an optimization flag then the real function
        will be returned without being wrapped.
    """
    if not __debug__:
        return func

    @_functools.wraps(func)
    def deprecated_function(*args, **kargs):
        _warnings.warn('This function name is deprecated',
                       DeprecationWarning, 2)
        return func(*args, **kargs)
    deprecated_function.__doc__ = None
    return deprecated_function

__all__ = []





def verify_order(order):
    order = order.upper()
    if order != 'C' and order != 'F':
        raise TypeError("order must be 'C' or 'F', not %r" % (order,))
    return order

def handle_order(shape, order):
    order = verify_order(order)
    if order == 'C':
        return shape
    else:
        return tuple(reversed(shape))



"""

Example::

    >>> import numpy as np
    >>> import tcod.path
    >>> dungeon = np.array(
    ...     [
    ...         [1, 0, 1, 1, 1],
    ...         [1, 0, 1, 0, 1],
    ...         [1, 1, 1, 0, 1],
    ...     ],
    ...     dtype=np.int8,
    ...     )
    ...
    >>> height, width = dungeon.shape

    # Create a pathfinder from a numpy array.
    # This is the recommended way to use the tcod.path module.
    >>> astar = tcod.path.AStar(dungeon)
    >>> print(astar.get_path(0, 0, 4, 2))
    [(0, 1), (1, 2), (2, 1), (3, 0), (4, 1), (4, 2)]
    >>> astar.cost[0, 1] = 1 # You can access the map array via this attribute.
    >>> print(astar.get_path(0, 0, 4, 2))
    [(1, 0), (2, 0), (3, 0), (4, 1), (4, 2)]

    # Create a pathfinder from an edge_cost function.
    # Calling Python functions from C is known to be very slow.
    >>> def edge_cost(my_x, my_y, dest_x, dest_y):
    ...     return dungeon[dest_y, dest_x]
    ...
    >>> dijkstra = tcod.path.Dijkstra(
    ...     tcod.path.EdgeCostCallback(edge_cost, width, height),
    ...     )
    ...
    >>> dijkstra.set_goal(0, 0)
    >>> print(dijkstra.get_path(4, 2))
    [(1, 0), (2, 0), (3, 0), (4, 1), (4, 2)]
"""

from __future__ import absolute_import

import numpy as np

from tcod.libtcod import lib, ffi


@ffi.def_extern()
def _pycall_path_old(x1, y1, x2, y2, handle):
    """libtcodpy style callback, needs to preserve the old userData issue."""
    func, userData = ffi.from_handle(handle)
    return func(x1, y1, x2, y2, userData)


@ffi.def_extern()
def _pycall_path_simple(x1, y1, x2, y2, handle):
    """Does less and should run faster, just calls the handle function."""
    return ffi.from_handle(handle)(x1, y1, x2, y2)


@ffi.def_extern()
def _pycall_path_swap_src_dest(x1, y1, x2, y2, handle):
    """A TDL function dest comes first to match up with a dest only call."""
    return ffi.from_handle(handle)(x2, y2, x1, y1)


@ffi.def_extern()
def _pycall_path_dest_only(x1, y1, x2, y2, handle):
    """A TDL function which samples the dest coordinate only."""
    return ffi.from_handle(handle)(x2, y2)

def _get_pathcost_func(name):
    """Return a properly cast PathCostArray callback."""
    return ffi.cast('TCOD_path_func_t', ffi.addressof(lib, name))


class _EdgeCostFunc(object):
    """
    Args:
        userdata (Any): Custom userdata to send to the C call.
        width (int): The maximum width of this callback.
        height (int): The maximum height of this callback.
    """
    _CALLBACK_P = lib._pycall_path_old

    def __init__(self, userdata, width, height):
        self._userdata = userdata
        self.width = width
        self.height = height

    def get_tcod_path_ffi(self):
        """

        Returns:
            Tuple[CData, CData]: A cffi (callback, userdata) pair.


        """
        return (self._CALLBACK_P, ffi.new_handle(self._userdata),
                self.width, self.height)

    def __repr__(self):
        return '%s(%r, width=%r, height=%r)' % (
            self.__class__.__name__,
            self._userdata, self.width, self.height,
            )


class EdgeCostCallback(_EdgeCostFunc):
    """Calculate cost from an edge-cost callback.

    Args:
        callback (Callable[[int, int, int, int], float]):
            A callback which can handle the following parameters:
            `(source_x:int, source_y:int, dest_x:int, dest_y:int) -> float`
        width (int): The maximum width of this callback.
        height (int): The maximum height of this callback.
    """
    _CALLBACK_P = lib._pycall_path_simple

    def __init__(self, callback, width, height):
        self.callback = callback
        super(EdgeCostCallback, self).__init__(callback, width, height)

class NodeCostArray(np.ndarray):
    """Calculate cost from a numpy array of nodes.

    Args:
        array (numpy.ndarray): A numpy array with the cost of each node.
    """

    _C_ARRAY_CALLBACKS = {
        np.float32: ('float*', _get_pathcost_func('PathCostArrayFloat32')),
        np.bool_: ('int8_t*', _get_pathcost_func('PathCostArrayInt8')),
        np.int8: ('int8_t*', _get_pathcost_func('PathCostArrayInt8')),
        np.uint8: ('uint8_t*', _get_pathcost_func('PathCostArrayUInt8')),
        np.int16: ('int16_t*', _get_pathcost_func('PathCostArrayInt16')),
        np.uint16: ('uint16_t*', _get_pathcost_func('PathCostArrayUInt16')),
        np.int32: ('int32_t*', _get_pathcost_func('PathCostArrayInt32')),
        np.uint32: ('uint32_t*', _get_pathcost_func('PathCostArrayUInt32')),
        }

    def __new__(cls, array):
        """Validate a numpy array and setup a C callback."""
        self = np.ascontiguousarray(array).view(cls)
        return self

    @property
    def width(self):
        return self.shape[1]

    @property
    def height(self):
        return self.shape[0]

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__,
                           repr(self.view(np.ndarray)))

    def get_tcod_path_ffi(self):
        if len(self.shape) != 2:
            raise ValueError('Array must have a 2d shape, shape is %r' %
                             (self.shape,))
        if self.dtype.type not in self._C_ARRAY_CALLBACKS:
            raise ValueError('dtype must be one of %r, dtype is %r' %
                             (self._C_ARRAY_CALLBACKS.keys(), self.dtype.type))

        array_type, callback = \
            self._C_ARRAY_CALLBACKS[self.dtype.type]
        userdata = ffi.new(
            'PathCostArray*',
            (self.width, ffi.cast(array_type, self.ctypes.data)),
        )
        return callback, userdata, self.width, self.height


class _PathFinder(object):
    """A class sharing methods used by AStar and Dijkstra."""

    def __init__(self, cost, diagonal=1.41):
        self.cost = cost
        self.diagonal = diagonal
        self._path_c = None
        self._callback = self._userdata = None

        if hasattr(self.cost, 'map_c'):
            self.width, self.height = self.cost.width, self.cost.height
            self._path_c = ffi.gc(
                self._path_new_using_map(self.cost.map_c, diagonal),
                self._path_delete,
                )
            return

        if not hasattr(self.cost, 'get_tcod_path_ffi'):
            assert not callable(self.cost), \
                "Any callback alone is missing width & height information. " \
                "Wrap your callback in tcod.path.EdgeCostCallback"
            self.cost = NodeCostArray(self.cost)

        self._callback, self._userdata, self.width, self.height = \
            self.cost.get_tcod_path_ffi()
        self._path_c = ffi.gc(
            self._path_new_using_function(
                self.width,
                self.height,
                self._callback,
                self._userdata,
                diagonal
                ),
            self._path_delete,
            )

    def __repr__(self):
        return '%s(cost=%r, diagonal=%r)' % (self.__class__.__name__,
                                             self.cost, self.diagonal)

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['_path_c']
        del state['width']
        del state['height']
        del state['_callback']
        del state['_userdata']
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)
        self.__init__(self.cost, self.diagonal)

    _path_new_using_map = lib.TCOD_path_new_using_map
    _path_new_using_function = lib.TCOD_path_new_using_function
    _path_delete = lib.TCOD_path_delete


class AStar(_PathFinder):
    """
    Args:
        cost (Union[tcod.map.Map, numpy.ndarray, Any]):
        diagonal (float): Multiplier for diagonal movement.
            A value of 0 will disable diagonal movement entirely.
    """

    def get_path(self, start_x, start_y, goal_x, goal_y):
        """Return a list of (x, y) steps to reach the goal point, if possible.

        Args:
            start_x (int): Starting X position.
            start_y (int): Starting Y position.
            goal_x (int): Destination X position.
            goal_y (int): Destination Y position.
        Returns:
            List[Tuple[int, int]]:
                A list of points, or an empty list if there is no valid path.
        """
        lib.TCOD_path_compute(self._path_c, start_x, start_y, goal_x, goal_y)
        path = []
        x = ffi.new('int[2]')
        y = x + 1
        while lib.TCOD_path_walk(self._path_c, x, y, False):
            path.append((x[0], y[0]))
        return path


class Dijkstra(_PathFinder):
    """
    Args:
        cost (Union[tcod.map.Map, numpy.ndarray, Any]):
        diagonal (float): Multiplier for diagonal movement.
            A value of 0 will disable diagonal movement entirely.
    """

    _path_new_using_map = lib.TCOD_dijkstra_new
    _path_new_using_function = lib.TCOD_dijkstra_new_using_function
    _path_delete = lib.TCOD_dijkstra_delete

    def set_goal(self, x, y):
        """Set the goal point and recompute the Dijkstra path-finder.
        """
        lib.TCOD_dijkstra_compute(self._path_c, x, y)

    def get_path(self, x, y):
        """Return a list of (x, y) steps to reach the goal point, if possible.
        """
        lib.TCOD_dijkstra_path_set(self._path_c, x, y)
        path = []
        pointer_x = ffi.new('int[2]')
        pointer_y = pointer_x + 1
        while lib.TCOD_dijkstra_path_walk(self._path_c, pointer_x, pointer_y):
            path.append((pointer_x[0], pointer_y[0]))
        return path



"""
    Getting Started
    ===============
      Once the library is imported you can load the font you want to use with
      :any:`tdl.set_font`.
      This is optional and when skipped will use a decent default font.

      After that you call :any:`tdl.init` to set the size of the window and
      get the root console in return.
      This console is the canvas to what will appear on the screen.

    Indexing Consoles
    =================
      For most methods taking a position you can use Python-style negative
      indexes to refer to the opposite side of a console with (-1, -1)
      starting at the bottom right.
      You can also check if a point is part of a console using containment
      logic i.e. ((x, y) in console).

      You may also iterate over a console using a for statement.  This returns
      every x,y coordinate available to draw on but it will be extremely slow
      to actually operate on every coordinate individualy.
      Try to minimize draws by using an offscreen :any:`Console`, only drawing
      what needs to be updated, and using :any:`Console.blit`.

    Drawing and Colors
    ==================

      Once you have the root console from :any:`tdl.init` you can start drawing
      on it using a method such as :any:`Console.draw_char`.
      When using this method you can have the char parameter be an integer or a
      single character string.

      The fg and bg parameters expect a variety of types.
      The parameters default to Ellipsis which will tell the function to
      use the colors previously set by the :any:`Console.set_colors` method.
      The colors set by :any`Console.set_colors` are per each
      :any:`Console`/:any:`Window` and default to white on black.
      You can use a 3-item list/tuple of [red, green, blue] with integers in
      the 0-255 range with [0, 0, 0] being black and [255, 255, 255] being
      white.
      You can even use a single integer of 0xRRGGBB if you like.

      Using None in the place of any of the three parameters (char, fg, bg)
      will tell the function to not overwrite that color or character.

      After the drawing functions are called a call to :any:`tdl.flush` will
      update the screen.
"""

from __future__ import (absolute_import, division,
                        print_function, unicode_literals)

import sys as _sys
import os as _os

import array as _array
import weakref as _weakref
import itertools as _itertools
import textwrap as _textwrap
import struct as _struct
import re as _re
import warnings as _warnings

from tcod import ffi as _ffi
from tcod import lib as _lib

import tdl.event
import tdl.noise
import tdl.map
import tdl.style as _style

try:
    from tdl.version import __version__
except ImportError: # Gets imported without version.py by ReadTheDocs
    __version__ = ''

_IS_PYTHON3 = (_sys.version_info[0] == 3)

if _IS_PYTHON3: # some type lists to use with isinstance
    _INTTYPES = (int,)
    _NUMTYPES = (int, float)
    _STRTYPES = (str, bytes)
else:
    _INTTYPES = (int, long)
    _NUMTYPES = (int, long, float)
    _STRTYPES = (str, unicode)

def _encodeString(string): # still used for filepaths, and that's about it
    "changes string into bytes if running in python 3, for sending to ctypes"
    if isinstance(string, _STRTYPES):
        return string.encode()
    return string

def _format_char(char):
    """Prepares a single character for passing to ctypes calls, needs to return
    an integer but can also pass None which will keep the current character
    instead of overwriting it.

    This is called often and needs to be optimized whenever possible.
    """
    if char is None:
        return -1
    if isinstance(char, _STRTYPES) and len(char) == 1:
        return ord(char)
    try:
        return int(char) # allow all int-like objects
    except:
        raise TypeError('char single character string, integer, or None\nReceived: ' + repr(char))

_utf32_codec = {'little': 'utf-32le', 'big': 'utf-32le'}[_sys.byteorder]

def _format_str(string):
    """Attempt fast string handing by decoding directly into an array."""
    if isinstance(string, _STRTYPES):
        if _IS_PYTHON3:
            array = _array.array('I')
            array.frombytes(string.encode(_utf32_codec))
        else: # Python 2
            if isinstance(string, unicode):
                array = _array.array(b'I')
                array.fromstring(string.encode(_utf32_codec))
            else:
                array = _array.array(b'B')
                array.fromstring(string)
        return array
    return string

_fontinitialized = False
_rootinitialized = False
_rootConsoleRef = None

_put_char_ex = _lib.TDL_console_put_char_ex

# python 2 to 3 workaround
if _sys.version_info[0] == 2:
    int_types = (int, long)
else:
    int_types = int


def _format_color(color, default=Ellipsis):
        if color is Ellipsis:
            return default
        if color is None:
            return -1
        if isinstance(color, (tuple, list)) and len(color) == 3:
            return (color[0] << 16) + (color[1] << 8) + color[2]
        try:
            return int(color) # allow all int-like objects
        except:
            raise TypeError('fg and bg must be a 3 item tuple, integer, Ellipsis, or None\nReceived: ' + repr(color))

def _to_tcod_color(color):
    return _ffi.new('TCOD_color_t *', (color >> 16 & 0xff,
                                       color >> 8 & 0xff,
                                       color & 0xff))

def _getImageSize(filename):
    """Try to get the width and height of a bmp of png image file"""
    result = None
    file = open(filename, 'rb')
    if file.read(8) == b'\x89PNG\r\n\x1a\n': # PNG
        while 1:
            length, = _struct.unpack('>i', file.read(4))
            chunkID = file.read(4)
            if chunkID == '': # EOF
                break
            if chunkID == b'IHDR':
                # return width, height
                result = _struct.unpack('>ii', file.read(8))
                break
            file.seek(4 + length, 1)
        file.close()
        return result
    file.seek(0)
    if file.read(8) == b'BM': # Bitmap
        file.seek(18, 0) # skip to size data
        result = _struct.unpack('<ii', file.read(8))
    file.close()
    return result # (width, height), or None

class TDLError(Exception):
    """
    The catch all for most TDL specific errors.
    """

class _BaseConsole(object):
    """You will not use this class directly.  The methods in this class are
    inherited by the :any:`tdl.Console` and :any:`tdl.Window` subclasses.
    """
    __slots__ = ('width', 'height', 'console', '_cursor', '_fg',
                 '_bg', '_blend', '__weakref__', '__dict__')

    def __init__(self):
        self._cursor = (0, 0)
        self._scrollMode = 'error'
        self._fg = _format_color((255, 255, 255))
        self._bg = _format_color((0, 0, 0))
        self._blend = _lib.TCOD_BKGND_SET

    def _normalizePoint(self, x, y):
        """Check if a point is in bounds and make minor adjustments.

        Respects Pythons negative indexes.  -1 starts at the bottom right.
        Replaces the _drawable function
        """
        # cast to int, always faster than type checking
        x = int(x)
        y = int(y)

        assert (-self.width <= x < self.width) and \
               (-self.height <= y < self.height), \
               ('(%i, %i) is an invalid postition on %s' % (x, y, self))

        # handle negative indexes
        return (x % self.width, y % self.height)

    def _normalizeRect(self, x, y, width, height):
        """Check if the rectangle is in bounds and make minor adjustments.
        raise AssertionError's for any problems
        """
        x, y = self._normalizePoint(x, y) # inherit _normalizePoint logic

        assert width is None or isinstance(width, _INTTYPES), 'width must be an integer or None, got %s' % repr(width)
        assert height is None or isinstance(height, _INTTYPES), 'height must be an integer or None, got %s' % repr(height)

        # if width or height are None then extend them to the edge
        if width is None:
            width = self.width - x
        elif width < 0: # handle negative numbers
            width += self.width
            width = max(0, width) # a 'too big' negative is clamped zero
        if height is None:
            height = self.height - y
            height = max(0, height)
        elif height < 0:
            height += self.height

        # reduce rect size to bounds
        width = min(width, self.width - x)
        height = min(height, self.height - y)

        return x, y, width, height

    def _normalizeCursor(self, x, y):
        """return the normalized the cursor position."""
        width, height = self.get_size()
        assert width != 0 and height != 0, 'can not print on a console with a width or height of zero'
        while x >= width:
            x -= width
            y += 1
        while y >= height:
            if self._scrollMode == 'scroll':
                y -= 1
                self.scroll(0, -1)
            elif self._scrollMode == 'error':
                # reset the cursor on error
                self._cursor = (0, 0)
                raise TDLError('Cursor has reached the end of the console')
        return (x, y)

    def set_mode(self, mode):
        """Configure how this console will react to the cursor writing past the
        end if the console.

        This is for methods that use the virtual cursor, such as
        :any:`print_str`.

        Args:
            mode (Text): The mode to set.

            Possible settings are:

           - 'error' - A TDLError will be raised once the cursor
             reaches the end of the console.  Everything up until
             the error will still be drawn.
             This is the default setting.
           - 'scroll' - The console will scroll up as stuff is
             written to the end.
             You can restrict the region with :any:`tdl.Window` when
             doing this.

        ..seealso:: :any:`write`, :any:`print_str`
        """
        MODES = ['error', 'scroll']
        if mode.lower() not in MODES:
            raise TDLError('mode must be one of %s, got %s' % (MODES, repr(mode)))
        self._scrollMode = mode.lower()

    def set_colors(self, fg=None, bg=None):
        """Sets the colors to be used with the L{print_str} and draw_* methods.

        Values of None will only leave the current values unchanged.

        Args:
            fg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])
            bg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])
        .. seealso:: :any:`move`, :any:`print_str`
        """
        if fg is not None:
            self._fg = _format_color(fg, self._fg)
        if bg is not None:
            self._bg = _format_color(bg, self._bg)

    def print_str(self, string):
        """Print a string at the virtual cursor.

        Handles special characters such as '\\n' and '\\r'.
        Printing past the bottom of the console will scroll everything upwards
        if :any:`set_mode` is set to 'scroll'.

        Colors can be set with :any:`set_colors` and the virtual cursor can
        be moved with :any:`move`.

        Args:
            string (Text): The text to print.

        .. seealso:: :any:`draw_str`, :any:`move`, :any:`set_colors`,
                     :any:`set_mode`, :any:`write`, :any:`Window`
        """
        x, y = self._cursor
        for char in string:
            if char == '\n': # line break
                x = 0
                y += 1
                continue
            if char == '\r': # return
                x = 0
                continue
            x, y = self._normalizeCursor(x, y)
            self.draw_char(x, y, char, self._fg, self._bg)
            x += 1
        self._cursor = (x, y)

    def write(self, string):
        """This method mimics basic file-like behaviour.

        Because of this method you can replace sys.stdout or sys.stderr with
        a :any:`Console` or :any:`Window` instance.

        This is a convoluted process and behaviour seen now can be excepted to
        change on later versions.

        Args:
            string (Text): The text to write out.

        .. seealso:: :any:`set_colors`, :any:`set_mode`, :any:`Window`
        """
        # some 'basic' line buffer stuff.
        # there must be an easier way to do this.  The textwrap module didn't
        # help much.
        x, y = self._normalizeCursor(*self._cursor)
        width, height = self.get_size()
        wrapper = _textwrap.TextWrapper(initial_indent=(' '*x), width=width)
        writeLines = []
        for line in string.split('\n'):
            if line:
                writeLines += wrapper.wrap(line)
                wrapper.initial_indent = ''
            else:
                writeLines.append([])

        for line in writeLines:
            x, y = self._normalizeCursor(x, y)
            self.draw_str(x, y, line[x:], self._fg, self._bg)
            y += 1
            x = 0
        y -= 1
        self._cursor = (x, y)

    def draw_char(self, x, y, char, fg=Ellipsis, bg=Ellipsis):
        """Draws a single character.

        Args:
            x (int): x-coordinate to draw on.
            y (int): y-coordinate to draw on.
            char (Optional[Union[int, Text]]): An integer, single character
                string, or None.

                You can set the char parameter as None if you only want to change
                the colors of the tile.
            fg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])
            bg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])

        Raises: AssertionError: Having x or y values that can't be placed
            inside of the console will raise an AssertionError.
            You can use always use ((x, y) in console) to
            check if a tile is drawable.

        .. seealso:: :any:`get_char`
        """
        #x, y = self._normalizePoint(x, y)
        _put_char_ex(self.console_c, x, y, _format_char(char),
                     _format_color(fg, self._fg), _format_color(bg, self._bg), 1)

    def draw_str(self, x, y, string, fg=Ellipsis, bg=Ellipsis):
        """Draws a string starting at x and y.

        A string that goes past the right side will wrap around.  A string
        wrapping to below the console will raise :any:`tdl.TDLError` but will
        still be written out.
        This means you can safely ignore the errors with a
        try..except block if you're fine with partially written strings.

        \\r and \\n are drawn on the console as normal character tiles.  No
        special encoding is done and any string will translate to the character
        table as is.

        For a string drawing operation that respects special characters see
        :any:`print_str`.

        Args:
            x (int): x-coordinate to start at.
            y (int): y-coordinate to start at.
            string (Union[Text, Iterable[int]]): A string or an iterable of
                numbers.

                Special characters are ignored and rendered as any other
                character.
            fg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])
            bg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])

        Raises:
            AssertionError: Having x or y values that can't be placed inside
                            of the console will raise an AssertionError.

                You can use always use ``((x, y) in console)`` to check if
                a tile is drawable.

        .. seealso:: :any:`print_str`
        """

        x, y = self._normalizePoint(x, y)
        fg, bg = _format_color(fg, self._fg), _format_color(bg, self._bg)
        width, height = self.get_size()
        def _drawStrGen(x=x, y=y, string=string, width=width, height=height):
            """Generator for draw_str

            Iterates over ((x, y), ch) data for _set_batch, raising an
            error if the end of the console is reached.
            """
            for char in _format_str(string):
                if y == height:
                    raise TDLError('End of console reached.')
                yield((x, y), char)
                x += 1 # advance cursor
                if x == width: # line break
                    x = 0
                    y += 1
        self._set_batch(_drawStrGen(), fg, bg)

    def draw_rect(self, x, y, width, height, string, fg=Ellipsis, bg=Ellipsis):
        """Draws a rectangle starting from x and y and extending to width and height.

        If width or height are None then it will extend to the edge of the console.

        Args:
            x (int): x-coordinate for the top side of the rect.
            y (int): y-coordinate for the left side of the rect.
            width (Optional[int]): The width of the rectangle.

                Can be None to extend to the bottom right of the
                console or can be a negative number to be sized reltive
                to the total size of the console.
            height (Optional[int]): The height of the rectangle.
            string (Optional[Union[Text, int]]): An integer, single character
                string, or None.

                You can set the string parameter as None if you only want
                to change the colors of an area.
            fg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])
            bg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])

        Raises:
            AssertionError: Having x or y values that can't be placed inside
                            of the console will raise an AssertionError.

            You can use always use ``((x, y) in console)`` to check if a tile
            is drawable.
        .. seealso:: :any:`clear`, :any:`draw_frame`
        """
        x, y, width, height = self._normalizeRect(x, y, width, height)
        fg, bg = _format_color(fg, self._fg), _format_color(bg, self._bg)
        char = _format_char(string)
        # use itertools to make an x,y grid
        # using ctypes here reduces type converstions later
        #grid = _itertools.product((_ctypes.c_int(x) for x in range(x, x + width)),
        #                          (_ctypes.c_int(y) for y in range(y, y + height)))
        grid = _itertools.product((x for x in range(x, x + width)),
                                  (y for y in range(y, y + height)))
        # zip the single character in a batch variable
        batch = zip(grid, _itertools.repeat(char, width * height))
        self._set_batch(batch, fg, bg, nullChar=(char is None))

    def draw_frame(self, x, y, width, height, string, fg=Ellipsis, bg=Ellipsis):
        """Similar to L{draw_rect} but only draws the outline of the rectangle.

        `width or `height` can be None to extend to the bottom right of the
        console or can be a negative number to be sized reltive
        to the total size of the console.

        Args:
            x (int): The x-coordinate to start on.
            y (int): The y-coordinate to start on.
            width (Optional[int]): Width of the rectangle.
            height (Optional[int]): Height of the rectangle.
            string (Optional[Union[Text, int]]): An integer, single character
                string, or None.

                You can set this parameter as None if you only want
                to change the colors of an area.
            fg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])
            bg (Optional[Union[Tuple[int, int, int], int, Ellipsis]])

        Raises:
            AssertionError: Having x or y values that can't be placed inside
                             of the console will raise an AssertionError.

             You can use always use ``((x, y) in console)`` to check if a tile
             is drawable.
        .. seealso:: :any:`draw_rect`, :any:`Window`
        """
        x, y, width, height = self._normalizeRect(x, y, width, height)
        fg, bg = _format_color(fg, self._fg), _format_color(bg, self._bg)
        char = _format_char(string)
        if width == 1 or height == 1: # it's just a single width line here
            return self.draw_rect(x, y, width, height, char, fg, bg)

        # draw sides of frame with draw_rect
        self.draw_rect(x, y, 1, height, char, fg, bg)
        self.draw_rect(x, y, width, 1, char, fg, bg)
        self.draw_rect(x + width - 1, y, 1, height, char, fg, bg)
        self.draw_rect(x, y + height - 1, width, 1, char, fg, bg)

    def blit(self, source, x=0, y=0, width=None, height=None, srcX=0, srcY=0,
             fg_alpha=1.0, bg_alpha=1.0):
        """Blit another console or Window onto the current console.

        By default it blits the entire source to the topleft corner.

        Args:
            source (Union[tdl.Console, tdl.Window]): The blitting source.
                A console can blit to itself without any problems.
            x (int): x-coordinate of this console to blit on.
            y (int): y-coordinate of this console to blit on.
            width (Optional[int]): Width of the rectangle.

                Can be None to extend as far as possible to the
                bottom right corner of the blit area or can be a negative
                number to be sized reltive to the total size of the
                B{destination} console.
            height (Optional[int]): Height of the rectangle.
            srcX (int):  x-coordinate of the source region to blit.
            srcY (int):  y-coordinate of the source region to blit.
            fg_alpha (float): The foreground alpha.
        """
        assert isinstance(source, (Console, Window)), "source muse be a Window or Console instance"

        # handle negative indexes and rects
        # negative width and height will be set realtive to the destination
        # and will also be clamped to the smallest Console
        x, y, width, height = self._normalizeRect(x, y, width, height)
        srcX, srcY, width, height = source._normalizeRect(srcX, srcY, width, height)

        # translate source and self if any of them are Window instances
        srcX, srcY = source._translate(srcX, srcY)
        source = source.console
        x, y = self._translate(x, y)
        self = self.console

        if self == source:
            # if we are the same console then we need a third console to hold
            # onto the data, otherwise it tries to copy into itself and
            # starts destroying everything
            tmp = Console(width, height)
            _lib.TCOD_console_blit(source.console_c,
                                   srcX, srcY, width, height,
                                   tmp.console_c, 0, 0, fg_alpha, bg_alpha)
            _lib.TCOD_console_blit(tmp.console_c, 0, 0, width, height,
                                   self.console_c, x, y, fg_alpha, bg_alpha)
        else:
            _lib.TCOD_console_blit(source.console_c,
                                   srcX, srcY, width, height,
                                   self.console_c, x, y, fg_alpha, bg_alpha)

    def get_cursor(self):
        """Return the virtual cursor position.

        The cursor can be moved with the :any:`move` method.

        Returns:
            Tuple[int, int]: The (x, y) coordinate of where :any:`print_str`
                will continue from.

        .. seealso:: :any:move`
        """
        x, y = self._cursor
        width, height = self.parent.get_size()
        while x >= width:
            x -= width
            y += 1
        if y >= height and self.scrollMode == 'scroll':
            y = height - 1
        return x, y

    def get_size(self):
        """Return the size of the console as (width, height)

        Returns:
            Tuple[int, int]: A (width, height) tuple.
        """
        return self.width, self.height

    def __iter__(self):
        """Return an iterator with every possible (x, y) value for this console.

        It goes without saying that working on the console this way is a
        slow process, especially for Python, and should be minimized.

        Returns:
            Iterator[Tuple[int, int]]: An ((x, y), ...) iterator.
        """
        return _itertools.product(range(self.width), range(self.height))

    def move(self, x, y):
        """Move the virtual cursor.

        Args:
            x (int): x-coordinate to place the cursor.
            y (int): y-coordinate to place the cursor.

        .. seealso:: :any:`get_cursor`, :any:`print_str`, :any:`write`
        """
        self._cursor = self._normalizePoint(x, y)

    def scroll(self, x, y):
        """Scroll the contents of the console in the direction of x,y.

        Uncovered areas will be cleared to the default background color.
        Does not move the virutal cursor.

        Args:
            x (int): Distance to scroll along the x-axis.
            y (int): Distance to scroll along the y-axis.

        Returns:
            Iterator[Tuple[int, int]]: An iterator over the (x, y) coordinates
            of any tile uncovered after scrolling.

        .. seealso:: :any:`set_colors`
        """
        assert isinstance(x, _INTTYPES), "x must be an integer, got %s" % repr(x)
        assert isinstance(y, _INTTYPES), "y must be an integer, got %s" % repr(x)
        def getSlide(x, length):
            """get the parameters needed to scroll the console in the given
            direction with x
            returns (x, length, srcx)
            """
            if x > 0:
                srcx = 0
                length -= x
            elif x < 0:
                srcx = abs(x)
                x = 0
                length -= srcx
            else:
                srcx = 0
            return x, length, srcx
        def getCover(x, length):
            """return the (x, width) ranges of what is covered and uncovered"""
            cover = (0, length) # everything covered
            uncover = None  # nothing uncovered
            if x > 0: # left side uncovered
                cover = (x, length - x)
                uncover = (0, x)
            elif x < 0: # right side uncovered
                x = abs(x)
                cover = (0, length - x)
                uncover = (length - x, x)
            return cover, uncover

        width, height = self.get_size()
        if abs(x) >= width or abs(y) >= height:
            return self.clear() # just clear the console normally

        # get the ranges of the areas that will be uncovered
        coverX, uncoverX = getCover(x, width)
        coverY, uncoverY = getCover(y, height)
        # so at this point we know that coverX and coverY makes a rect that
        # encases the area that we end up blitting to.  uncoverX/Y makes a
        # rect in the corner of the uncovered area.  So we need to combine
        # the uncoverX/Y with coverY/X to make what's left of the uncovered
        # area.  Explaining it makes it mush easier to do now.

        # But first we need to blit.
        x, width, srcx = getSlide(x, width)
        y, height, srcy = getSlide(y, height)
        self.blit(self, x, y, width, height, srcx, srcy)
        if uncoverX: # clear sides (0x20 is space)
            self.draw_rect(uncoverX[0], coverY[0], uncoverX[1], coverY[1],
                           0x20, self._fg, self._bg)
        if uncoverY: # clear top/bottom
            self.draw_rect(coverX[0], uncoverY[0], coverX[1], uncoverY[1],
                           0x20, self._fg, self._bg)
        if uncoverX and uncoverY: # clear corner
            self.draw_rect(uncoverX[0], uncoverY[0], uncoverX[1], uncoverY[1],
                           0x20, self._fg, self._bg)

    def clear(self, fg=Ellipsis, bg=Ellipsis):
        """Clears the entire L{Console}/L{Window}.

        Unlike other drawing functions, fg and bg can not be None.

        Args:
            fg (Union[Tuple[int, int, int], int, Ellipsis])
            bg (Union[Tuple[int, int, int], int, Ellipsis])

        .. seealso:: :any:`draw_rect`
        """
        raise NotImplementedError('this method is overwritten by subclasses')

    def get_char(self, x, y):
        """Return the character and colors of a tile as (ch, fg, bg)

        This method runs very slowly as is not recommended to be called
        frequently.

        Args:
            x (int): The x-coordinate to pick.
            y (int): The y-coordinate to pick.

        Returns:
            Tuple[int, Tuple[int, int, int], Tuple[int, int, int]]: A 3-item
                tuple: `(int, fg, bg)`

                The first item is an integer of the
                character at the position (x, y) the second and third are the
                foreground and background colors respectfully.

        .. seealso:: :any:`draw_char`
        """
        raise NotImplementedError('Method here only exists for the docstring')

    def __contains__(self, position):
        """Use ``((x, y) in console)`` to check if a position is drawable on
        this console.
        """
        x, y = position
        return (0 <= x < self.width) and (0 <= y < self.height)

class Console(_BaseConsole):
    """Contains character and color data and can be drawn to.

    The console created by the :any:`tdl.init` function is the root console
    and is the console that is rendered to the screen with :any:`flush`.

    Any console created from the Console class is an off-screen console that
    can be drawn on before being :any:`blit` to the root console.

    Args:
        width (int): Width of the new console in tiles
        height (int): Height of the new console in tiles
    """

    def __init__(self, width, height):
        _BaseConsole.__init__(self)
        self.console_c = _lib.TCOD_console_new(width, height)
        self.console = self
        self.width = width
        self.height = height

    @property
    def tcod_console(self):
        return self.console_c
    @tcod_console.setter
    def tcod_console(self, value):
        self.console_c = value

    @classmethod
    def _newConsole(cls, console):
        """Make a Console instance, from a console ctype"""
        self = cls.__new__(cls)
        _BaseConsole.__init__(self)
        self.console_c = console
        self.console = self
        self.width = _lib.TCOD_console_get_width(console)
        self.height = _lib.TCOD_console_get_height(console)
        return self

    def _root_unhook(self):
        """Change this root console into a normal Console object and
        delete the root console from TCOD
        """
        global _rootinitialized, _rootConsoleRef
        # do we recognise this as the root console?
        # if not then assume the console has already been taken care of
        if(_rootConsoleRef and _rootConsoleRef() is self):
            # turn this console into a regular console
            unhooked = _lib.TCOD_console_new(self.width, self.height)
            _lib.TCOD_console_blit(self.console_c,
                                   0, 0, self.width, self.height,
                                   unhooked, 0, 0, 1, 1)
            # delete root console from TDL and TCOD
            _rootinitialized = False
            _rootConsoleRef = None
            _lib.TCOD_console_delete(self.console_c)
            # this Console object is now a regular console
            self.console_c = unhooked

    def __del__(self):
        """
        If the main console is garbage collected then the window will be closed as well
        """
        if self.console_c is None:
            return # this console was already deleted
        if self.console_c is _ffi.NULL:
            # a pointer to the special root console
            self._root_unhook() # unhook the console and leave it to the GC
            return
        # this is a normal console pointer and can be safely deleted
        _lib.TCOD_console_delete(self.console_c)
        self.console_c = None

    def __copy__(self):
        # make a new class and blit
        clone = self.__class__(self.width, self.height)
        clone.blit(self)
        return clone

    def __getstate__(self):
        # save data from get_char
        data = [self.get_char(x, y) for x,y in
                _itertools.product(range(self.width), range(self.height))]
        return self.width, self.height, data

    def __setstate__(self, state):
        # make console from __init__ and unpack a get_char array
        width, height, data = state
        self.__init__(width, height)
        for (x, y), graphic in zip(_itertools.product(range(width),
                                                      range(height)), data):
            self.draw_char(x, y, *graphic)

    @staticmethod
    def _translate(x, y):
        """Convertion x and y to their position on the root Console for this Window

        Because this is a Console instead of a Window we return the paramaters
        untouched"""
        return x, y

    def clear(self, fg=Ellipsis, bg=Ellipsis):
        # inherit docstring
        assert fg is not None and bg is not None, 'Can not use None with clear'
        fg = _format_color(fg, self._fg)
        bg = _format_color(bg, self._bg)
        _lib.TCOD_console_set_default_foreground(self.console_c,
                                                 _to_tcod_color(fg)[0])
        _lib.TCOD_console_set_default_background(self.console_c,
                                                 _to_tcod_color(bg)[0])
        _lib.TCOD_console_clear(self.console_c)


    def _set_char(self, x, y, char, fg=None, bg=None,
                  bgblend=_lib.TCOD_BKGND_SET):
        """
        Sets a character.
        This is called often and is designed to be as fast as possible.

        Because of the need for speed this function will do NO TYPE CHECKING
        AT ALL, it's up to the drawing functions to use the functions:
        _format_char and _format_color before passing to this."""
        # values are already formatted, honestly this function is redundant
        return _put_char_ex(self.console_c, x, y, char, fg, bg, bgblend)

    def _set_batch(self, batch, fg, bg, bgblend=1, nullChar=False):
        """
        Try to perform a batch operation otherwise fall back to _set_char.
        If fg and bg are defined then this is faster but not by very
        much.

        if any character is None then nullChar is True

        batch is a iterable of [(x, y), ch] items
        """
        for (x, y), char in batch:
            self._set_char(x, y, char, fg, bg, bgblend)

    def get_char(self, x, y):
        # inherit docstring
        x, y = self._normalizePoint(x, y)
        char = _lib.TCOD_console_get_char(self.console_c, x, y)
        bg = _lib.TCOD_console_get_char_background(self.console_c, x, y)
        fg = _lib.TCOD_console_get_char_foreground(self.console_c, x, y)
        return char, (fg.r, fg.g, fg.b), (bg.r, bg.g, bg.b)

    # Copy docstrings for Sphinx
    clear.__doc__ = _BaseConsole.clear.__doc__
    get_char.__doc__ = _BaseConsole.get_char.__doc__

    def __repr__(self):
        return "<Console (Width=%i Height=%i)>" % (self.width, self.height)


class Window(_BaseConsole):
    """Isolate part of a :any:`Console` or :any:`Window` instance.

    This classes methods are the same as :any:`tdl.Console`

    Making a Window and setting its width or height to None will extend it to
    the edge of the console.

    This follows the normal rules for indexing so you can use a
    negative integer to place the Window relative to the bottom
    right of the parent Console instance.

    `width` or `height` can be set to None to extend as far as possible
    to the bottom right corner of the parent Console or can be a
    negative number to be sized reltive to the Consoles total size.

    Args:
        console (Union(tdl.Console, tdl.Window)): The parent object.
        x (int): x-coordinate to place the Window.
        y (int): y-coordinate to place the Window.
        width (Optional[int]): Width of the Window.
        height (Optional[int]): Height of the Window.
    """

    __slots__ = ('parent', 'x', 'y')

    def __init__(self, console, x, y, width, height):
        _BaseConsole.__init__(self)
        assert isinstance(console, (Console, Window)), 'console parameter must be a Console or Window instance, got %s' % repr(console)
        self.parent = console
        self.x, self.y, self.width, self.height = console._normalizeRect(x, y, width, height)
        if isinstance(console, Console):
            self.console = console
        else:
            self.console = self.parent.console

    def _translate(self, x, y):
        """Convertion x and y to their position on the root Console"""
        # we add our position relative to our parent and then call then next parent up
        return self.parent._translate((x + self.x), (y + self.y))

    def clear(self, fg=Ellipsis, bg=Ellipsis):
        # inherit docstring
        assert fg is not None and bg is not None, 'Can not use None with clear'
        if fg is Ellipsis:
            fg = self._fg
        if bg is Ellipsis:
            bg = self._bg
        self.draw_rect(0, 0, None, None, 0x20, fg, bg)

    def _set_char(self, x, y, char=None, fg=None, bg=None, bgblend=1):
        self.parent._set_char((x + self.x), (y + self.y), char, fg, bg, bgblend)

    def _set_batch(self, batch, *args, **kargs):
        # positional values will need to be translated to the parent console
        myX = self.x # remove dots for speed up
        myY = self.y
        self.parent._set_batch((((x + myX, y + myY), ch)
                                   for ((x, y), ch) in batch), *args, **kargs)


    def draw_char(self, x, y, char, fg=Ellipsis, bg=Ellipsis):
        # inherit docstring
        x, y = self._normalizePoint(x, y)
        if fg is Ellipsis:
            fg = self._fg
        if bg is Ellipsis:
            bg = self._bg
        self.parent.draw_char(x + self.x, y + self.y, char, fg, bg)

    def draw_rect(self, x, y, width, height, string, fg=Ellipsis, bg=Ellipsis):
        # inherit docstring
        x, y, width, height = self._normalizeRect(x, y, width, height)
        if fg is Ellipsis:
            fg = self._fg
        if bg is Ellipsis:
            bg = self._bg
        self.parent.draw_rect(x + self.x, y + self.y, width, height,
                              string, fg, bg)

    def draw_frame(self, x, y, width, height, string, fg=Ellipsis, bg=Ellipsis):
        # inherit docstring
        x, y, width, height = self._normalizeRect(x, y, width, height)
        if fg is Ellipsis:
            fg = self._fg
        if bg is Ellipsis:
            bg = self._bg
        self.parent.draw_frame(x + self.x, y + self.y, width, height,
                               string, fg, bg)

    def get_char(self, x, y):
        # inherit docstring
        x, y = self._normalizePoint(x, y)
        xtrans, ytrans = self._translate(x, y)
        return self.console.get_char(xtrans, ytrans)

    def __repr__(self):
        return "<Window(X=%i Y=%i Width=%i Height=%i)>" % (self.x, self.y,
                                                          self.width,
                                                          self.height)


def init(width, height, title=None, fullscreen=False, renderer='SDL'):
    """Start the main console with the given width and height and return the
    root console.

    Call the consoles drawing functions.  Then remember to use L{tdl.flush} to
    make what's drawn visible on the console.

    Args:
        width (int): width of the root console (in tiles)
        height (int): height of the root console (in tiles)
        title (Optiona[Text]):
            Text to display as the window title.

            If left None it defaults to the running scripts filename.
        fullscreen (bool): Can be set to True to start in fullscreen mode.
        renderer (Text): Can be one of 'GLSL', 'OPENGL', or 'SDL'.

        Due to way Python works you're unlikely to see much of an
        improvement by using 'GLSL' over 'OPENGL' as most of the
        time Python is slow interacting with the console and the
        rendering itself is pretty fast even on 'SDL'.

    Returns:
        tdl.Console: The root console.

            Only what is drawn on the root console is
            what's visible after a call to :any:`tdl.flush`.
            After the root console is garbage collected, the window made by
            this function will close.

    .. seealso::
        :any:`Console` :any:`set_font`
    """
    RENDERERS = {'GLSL': 0, 'OPENGL': 1, 'SDL': 2}
    global _rootinitialized, _rootConsoleRef
    if not _fontinitialized: # set the default font to the one that comes with tdl
        set_font(_os.path.join(__path__[0], 'terminal8x8.png'),
                 None, None, True, True)

    if renderer.upper() not in RENDERERS:
        raise TDLError('No such render type "%s", expected one of "%s"' % (renderer, '", "'.join(RENDERERS)))
    renderer = RENDERERS[renderer.upper()]

    # If a console already exists then make a clone to replace it
    if _rootConsoleRef and _rootConsoleRef():
        # unhook the root console, turning into a regular console and deleting
        # the root console from libTCOD
        _rootConsoleRef()._root_unhook()

    if title is None: # use a default title
        if _sys.argv:
            # Use the script filename as the title.
            title = _os.path.basename(_sys.argv[0])
        else:
            title = 'python-tdl'

    _lib.TCOD_console_init_root(width, height, _encodeString(title), fullscreen, renderer)

    #event.get() # flush the libtcod event queue to fix some issues
    # issues may be fixed already

    event._eventsflushed = False
    _rootinitialized = True
    rootconsole = Console._newConsole(_ffi.NULL)
    _rootConsoleRef = _weakref.ref(rootconsole)

    return rootconsole

def flush():
    """Make all changes visible and update the screen.

    Remember to call this function after drawing operations.
    Calls to flush will enfore the frame rate limit set by :any:`tdl.set_fps`.

    This function can only be called after :any:`tdl.init`
    """
    if not _rootinitialized:
        raise TDLError('Cannot flush without first initializing with tdl.init')
    # flush the OS event queue, preventing lock-ups if not done manually
    event.get()
    _lib.TCOD_console_flush()

def set_font(path, columns=None, rows=None, columnFirst=False,
             greyscale=False, altLayout=False):
    """Changes the font to be used for this session.
    This should be called before :any:`tdl.init`

    If the font specifies its size in its filename (i.e. font_NxN.png) then this
    function can auto-detect the tileset formatting and the parameters columns
    and rows can be left None.

    While it's possible you can change the font mid program it can sometimes
    break in rare circumstances.  So use caution when doing this.

    Args:
        path (Text): A file path to a `.bmp` or `.png` file.
        columns (int):
            Number of columns in the tileset.

            Can be left None for auto-detection.
        rows (int):
            Number of rows in the tileset.

            Can be left None for auto-detection.
        columnFirst (bool):
            Defines if the characer order goes along the rows or colomns.

            It should be True if the charater codes 0-15 are in the
            first column, and should be False if the characters 0-15
            are in the first row.
        greyscale (bool):
            Creates an anti-aliased font from a greyscale bitmap.
            Otherwise it uses the alpha channel for anti-aliasing.

            Unless you actually need anti-aliasing from a font you
            know uses a smooth greyscale channel you should leave
            this on False.
        altLayout (bool)
            An alternative layout with space in the upper left corner.
            The colomn parameter is ignored if this is True,
            find examples of this layout in the `font/libtcod/`
            directory included with the python-tdl source.

    Raises:
        TDLError: Will be raised if no file is found at path or if auto-
                  detection fails.
    """
    # put up some constants that are only used here
    FONT_LAYOUT_ASCII_INCOL = 1
    FONT_LAYOUT_ASCII_INROW = 2
    FONT_TYPE_GREYSCALE = 4
    FONT_LAYOUT_TCOD = 8
    global _fontinitialized
    _fontinitialized = True
    flags = 0
    if altLayout:
        flags |= FONT_LAYOUT_TCOD
    elif columnFirst:
        flags |= FONT_LAYOUT_ASCII_INCOL
    else:
        flags |= FONT_LAYOUT_ASCII_INROW
    if greyscale:
        flags |= FONT_TYPE_GREYSCALE
    if not _os.path.exists(path):
        raise TDLError('Font %r not found.' % _os.path.abspath(path))
    path = _os.path.abspath(path)

    # and the rest is the auto-detect script
    imgSize = _getImageSize(path) # try to find image size
    if imgSize:
        fontWidth, fontHeight = None, None
        imgWidth, imgHeight = imgSize
        # try to get font size from filename
        match = _re.match('.*?([0-9]+)[xX]([0-9]+)', _os.path.basename(path))
        if match:
            fontWidth, fontHeight = match.groups()
            fontWidth, fontHeight = int(fontWidth), int(fontHeight)

            # estimate correct tileset size
            estColumns, remC = divmod(imgWidth, fontWidth)
            estRows, remR = divmod(imgHeight, fontHeight)
            if remC or remR:
                _warnings.warn("Font may be incorrectly formatted.")

            if not columns:
                columns = estColumns
            if not rows:
                rows = estRows
        else:
            # filename doesn't contain NxN, but we can still estimate the fontWidth
            # and fontHeight given number of columns and rows.
            if columns and rows:
                fontWidth, remC = divmod(imgWidth, columns)
                fontHeight, remR = divmod(imgHeight, rows)
                if remC or remR:
                    _warnings.warn("Font may be incorrectly formatted.")

            # the font name excluded the fonts size
            if not (columns and rows):
                # no matched font size and no tileset is given
                raise TDLError('%s has no font size in filename' % _os.path.basename(path))

        if columns and rows:
            # confirm user set options
            if (fontWidth * columns != imgWidth or
                fontHeight * rows != imgHeight):
                _warnings.warn("set_font parameters are set as if the image size is (%d, %d) when the detected size is actually (%i, %i)"
                             % (fontWidth * columns, fontHeight * rows,
                                imgWidth, imgHeight))
    else:
        _warnings.warn("%s is probably not an image." % _os.path.basename(path))

    if not (columns and rows):
        # didn't auto-detect
        raise TDLError('Can not auto-detect the tileset of %s' % _os.path.basename(path))

    _lib.TCOD_console_set_custom_font(_encodeString(path), flags, columns, rows)

def get_fullscreen():
    """Returns True if program is fullscreen.

    Returns:
        bool:  Returns True if the application is in full-screen mode.
               Otherwise returns False.
    """
    if not _rootinitialized:
        raise TDLError('Initialize first with tdl.init')
    return _lib.TCOD_console_is_fullscreen()

def set_fullscreen(fullscreen):
    """Changes the fullscreen state.

    Args:
        fullscreen (bool): True for full-screen, False for windowed mode.
    """
    if not _rootinitialized:
        raise TDLError('Initialize first with tdl.init')
    _lib.TCOD_console_set_fullscreen(fullscreen)

def set_title(title):
    """Change the window title.

    Args:
        title (Text): The new title text.
    """
    if not _rootinitialized:
        raise TDLError('Not initilized.  Set title with tdl.init')
    _lib.TCOD_console_set_window_title(_encodeString(title))

def screenshot(path=None):
    """Capture the screen and save it as a png file.

    If path is None then the image will be placed in the current
    folder with the names:
    ``screenshot001.png, screenshot002.png, ...``

    Args:
        path (Optional[Text]): The file path to save the screenshot.
    """
    if not _rootinitialized:
        raise TDLError('Initialize first with tdl.init')
    if isinstance(path, str):
        _lib.TCOD_sys_save_screenshot(_encodeString(path))
    elif path is None: # save to screenshot001.png, screenshot002.png, ...
        filelist = _os.listdir('.')
        n = 1
        filename = 'screenshot%.3i.png' % n
        while filename in filelist:
            n += 1
            filename = 'screenshot%.3i.png' % n
        _lib.TCOD_sys_save_screenshot(_encodeString(filename))
    else: # assume file like obj
        #save to temp file and copy to file-like obj
        tmpname = _os.tempnam()
        _lib.TCOD_sys_save_screenshot(_encodeString(tmpname))
        with tmpname as tmpfile:
            path.write(tmpfile.read())
        _os.remove(tmpname)
    #else:
    #    raise TypeError('path is an invalid type: %s' % type(path))

def set_fps(fps):
    """Set the maximum frame rate.

    Further calls to :any:`tdl.flush` will limit the speed of
    the program to run at `fps` frames per second. This can
    also be set to None to remove the frame rate limit.

    Args:
        fps (optional[int]): The frames per second limit, or None.
    """
    _lib.TCOD_sys_set_fps(fps or 0)

def get_fps():
    """Return the current frames per second of the running program set by
    :any:`set_fps`

    Returns:
        int: The frame rate set by :any:`set_fps`.
             If there is no current limit, this will return 0.
    """
    return _lib.TCOD_sys_get_fps()

def force_resolution(width, height):
    """Change the fullscreen resoulution.

    Args:
        width (int): Width in pixels.
        height (int): Height in pixels.
    """
    _lib.TCOD_sys_force_fullscreen_resolution(width, height)


__all__ = [_var for _var in locals().keys() if _var[0] != '_'] # remove modules from __all__
__all__ += ['_BaseConsole'] # keep this object public to show the documentation in epydoc
__all__.remove('absolute_import')
__all__.remove('division')
__all__.remove('print_function')
__all__.remove('unicode_literals')

# backported function names
_BaseConsole.setMode = _style.backport(_BaseConsole.set_mode)
_BaseConsole.setColors = _style.backport(_BaseConsole.set_colors)
_BaseConsole.printStr = _style.backport(_BaseConsole.print_str)
_BaseConsole.drawChar = _style.backport(_BaseConsole.draw_char)
_BaseConsole.drawStr = _style.backport(_BaseConsole.draw_str)
_BaseConsole.drawRect = _style.backport(_BaseConsole.draw_rect)
_BaseConsole.drawFrame = _style.backport(_BaseConsole.draw_frame)
_BaseConsole.getCursor = _style.backport(_BaseConsole.get_cursor)
_BaseConsole.getSize = _style.backport(_BaseConsole.get_size)
_BaseConsole.getChar = _style.backport(_BaseConsole.get_char)

Console.getChar = _style.backport(Console.get_char)

Window.drawChar = _style.backport(Window.draw_char)
Window.drawRect = _style.backport(Window.draw_rect)
Window.drawFrame = _style.backport(Window.draw_frame)
Window.getChar = _style.backport(Window.get_char)

setFont = _style.backport(set_font)
getFullscreen = _style.backport(get_fullscreen)
setFullscreen = _style.backport(set_fullscreen)
setTitle = _style.backport(set_title)
setFPS = _style.backport(set_fps)
getFPS = _style.backport(get_fps)
forceResolution = _style.backport(force_resolution)



'''
The following example shows how to travrse the BSP tree using Python.  This
assumes `create_room` and `connect_rooms` will be replaced by custom code.

Example::

    import tcod.bsp

    def create_room(node):
        """Initialize the room at this current node."""
        print('Create a room for %s.' % node)

    def connect_rooms(node):
        """Connect two fully initialized rooms."""
        node1, node2 = node.children
        print('Connect the rooms:\\n%s\\n%s' % (node1, node2))

    def traverse(node):
        """Traverse a BSP tree dispatching nodes to the correct calls."""
        # For nodes without children, node.children is an empty tuple.
        for child in node.children:
            traverse(child)

        if node.children:
            connect_rooms(node)
        else:
            create_room(node)

    bsp = tcod.bsp.BSP(x=0, y=0, width=80, height=60)
    bsp.split_recursive(
        depth=5,
        min_width=3,
        min_height=3,
        max_horizontal_ratio=1.5,
        max_vertical_ratio=1.5,
        )
    traverse(bsp)
'''
from __future__ import absolute_import as _

from tcod.libtcod import lib, ffi


class BSP(object):
    """A binary space partitioning tree which can be used for simple dungeon
    generation.

    Attributes:
        x (int): Rectangle left coordinate.
        y (int): Rectangle top coordinate.
        width (int): Rectangle width.
        height (int): Rectangle height.
        level (int): This nodes depth.
        position (int): The integer of where the node was split.
        horizontal (bool): This nodes split orientation.
        parent (Optional[BSP]): This nodes parent or None
        children (Optional[Tuple[BSP, BSP]]):
            A tuple of (left, right) BSP instances, or
            None if this BSP has no children.

    Args:
        x (int): Rectangle left coordinate.
        y (int): Rectangle top coordinate.
        width (int): Rectangle width.
        height (int): Rectangle height.
    """

    def __init__(self, x, y, width, height):
        self.x = x
        self.y = y
        self.width = width
        self.height = height

        self.level = 0
        self.position = 0
        self.horizontal = False

        self.parent = None
        self.children = ()

    @property
    def w(self):
        return self.width
    @w.setter
    def w(self, value):
        self.width = value

    @property
    def h(self):
        return self.height
    @h.setter
    def h(self, value):
        self.height = value

    def _as_cdata(self):
        cdata = ffi.gc(lib.TCOD_bsp_new_with_size(self.x, self.y,
                                                  self.width, self.height),
                       lib.TCOD_bsp_delete)
        cdata.level = self.level
        return cdata

    def __str__(self):
        """Provide a useful readout when printed."""
        status = 'leaf'
        if self.children:
            status = ('split at position=%i,horizontal=%r' %
                      (self.position, self.horizontal))

        return ('<%s(x=%i,y=%i,width=%i,height=%i)level=%i,%s>' %
                (self.__class__.__name__,
                 self.x, self.y, self.width, self.height, self.level, status))

    def _unpack_bsp_tree(self, cdata):
        self.x = cdata.x
        self.y = cdata.y
        self.width = cdata.w
        self.height = cdata.h
        self.level = cdata.level
        self.position = cdata.position
        self.horizontal = bool(cdata.horizontal)
        if lib.TCOD_bsp_is_leaf(cdata):
            return
        self.children = (BSP(0, 0, 0, 0), BSP(0, 0, 0, 0))
        self.children[0].parent = self
        self.children[0]._unpack_bsp_tree(lib.TCOD_bsp_left(cdata))
        self.children[1].parent = self
        self.children[1]._unpack_bsp_tree(lib.TCOD_bsp_right(cdata))

    def split_once(self, horizontal, position):
        """Split this partition into 2 sub-partitions.

        Args:
            horizontal (bool):
            position (int):
        """
        cdata = self._as_cdata()
        lib.TCOD_bsp_split_once(cdata, horizontal, position)
        self._unpack_bsp_tree(cdata)

    def split_recursive(self, depth, min_width, min_height,
                        max_horizontal_ratio, max_vertical_ratio, seed=None):
        """Divide this partition recursively.

        Args:
            depth (int): The maximum depth to divide this object recursively.
            min_width (int): The minimum width of any individual partition.
            min_height (int): The minimum height of any individual partition.
            max_horizontal_ratio (float):
                Prevent creating a horizontal ratio more extreme than this.
            max_vertical_ratio (float):
                Prevent creating a vertical ratio more extreme than this.
            seed (Optional[tcod.random.Random]):
                The random number generator to use.
        """
        cdata = self._as_cdata()
        lib.TCOD_bsp_split_recursive(
            cdata,
            seed or ffi.NULL,
            depth,
            min_width, min_height,
            max_horizontal_ratio, max_vertical_ratio,
        )
        self._unpack_bsp_tree(cdata)

    def walk(self):
        """Iterate over this BSP's hieracrhy.

        The iterator will include the instance which called it.
        It will traverse its own children and grandchildren, in no particular
        order.

        Returns:
            Iterator[BSP]: An iterator of BSP nodes.

        .. deprecated:: 2.3
            See the module example for how to iterate over a BSP tree.
        """
        return self._iter_post_order()

    def _iter_pre_order(self):
        yield self
        for child in self.children:
            for grandchild in child._iter_pre_order():
                yield grandchild

    def _iter_in_order(self):
        if self.children:
            for grandchild in self.children[0]._iter_in_order():
                yield grandchild
            yield self
            for grandchild in self.children[1]._iter_in_order():
                yield grandchild
        else:
            yield self

    def _iter_post_order(self):
        for child in self.children:
            for grandchild in child._iter_post_order():
                yield grandchild
        yield self

    def _iter_level_order(self):
        return sorted(self._iter_pre_order(), key=lambda n:n.level)

    def _iter_inverted_level_order(self):
        return reversed(self._iter_level_order())

    def contains(self, x, y):
        """Returns True if this node contains these coordinates.

        Args:
            x (int): X position to check.
            y (int): Y position to check.

        Returns:
            bool: True if this node contains these coordinates.
                  Otherwise False.
        """
        return (self.x <= x < self.x + self.width and
                self.y <= y < self.y + self.height)

    def find_node(self, x, y):
        """Return the deepest node which contains these coordinates.

        Returns:
            Optional[BSP]: BSP object or None.
        """
        if not self.contains(x, y):
            return None
        for child in self.children:
            found = child.find_node(x, y)
            if found:
                return found
        return self



"""
The :any:`Noise.sample_mgrid` and :any:`Noise.sample_ogrid` methods are
multi-threaded operations when the Python runtime supports OpenMP.
Even when single threaded these methods will perform much better than
multiple calls to :any:`Noise.get_point`.

Example::

    import numpy as np
    import tcod
    import tcod.noise

    noise = tcod.noise.Noise(
        dimensions=2,
        algorithm=tcod.NOISE_SIMPLEX,
        implementation=tcod.noise.TURBULENCE,
        hurst=0.5,
        lacunarity=2.0,
        octaves=4,
        seed=None,
        )

    # Create a 5x5 open multi-dimensional mesh-grid.
    ogrid = [np.arange(5, dtype=np.float32),
             np.arange(5, dtype=np.float32)]
    print(ogrid)

    # Scale the grid.
    ogrid[0] *= 0.25
    ogrid[1] *= 0.25

    # Return the sampled noise from this grid of points.
    samples = noise.sample_ogrid(ogrid)
    print(samples)
"""
from __future__ import absolute_import

import numpy as np

from tcod.libtcod import ffi, lib
import tcod.libtcod

"""Noise implementation constants"""
SIMPLE = 0
FBM = 1
TURBULENCE = 2

class Noise(object):
    """

    The ``hurst`` exponent describes the raggedness of the resultant noise,
    with a higher value leading to a smoother noise.
    Not used with tcod.noise.SIMPLE.

    ``lacunarity`` is a multiplier that determines how fast the noise
    frequency increases for each successive octave.
    Not used with tcod.noise.SIMPLE.

    Args:
        dimensions (int): Must be from 1 to 4.
        algorithm (int): Defaults to NOISE_SIMPLEX
        implementation (int): Defaults to tcod.noise.SIMPLE
        hurst (float): The hurst exponent.  Should be in the 0.0-1.0 range.
        lacunarity (float): The noise lacunarity.
        octaves (float): The level of detail on fBm and turbulence
                         implementations.
        seed (Optional[Random]): A Random instance, or None.

    Attributes:
        noise_c (CData): A cffi pointer to a TCOD_noise_t object.
    """

    def __init__(self, dimensions, algorithm=2, implementation=SIMPLE,
                 hurst=0.5, lacunarity=2.0, octaves=4, seed=None):
        if not 0 < dimensions <= 4:
            raise ValueError('dimensions must be in range 0 < n <= 4, got %r' %
                             (dimensions,))
        self._random = seed
        _random_c = seed.random_c if seed else ffi.NULL
        self._algorithm = algorithm
        self.noise_c = ffi.gc(
            ffi.cast(
                'perlin_data_t*',
                lib.TCOD_noise_new(dimensions, hurst, lacunarity,
                                   _random_c),
                ),
            lib.TCOD_noise_delete)
        self._tdl_noise_c = ffi.new('TDLNoise*', (self.noise_c,
                                                  dimensions,
                                                  0,
                                                  octaves))
        self.implementation = implementation # sanity check

    @property
    def dimensions(self):
        return self._tdl_noise_c.dimensions

    @property
    def dimentions(self): # deprecated
        return self.dimensions

    @property
    def algorithm(self):
        return self.noise_c.noise_type
    @algorithm.setter
    def algorithm(self, value):
        lib.TCOD_noise_set_type(self.noise_c, value)

    @property
    def implementation(self):
        return self._tdl_noise_c.implementation
    @implementation.setter
    def implementation(self, value):
        if not 0 <= value < 3:
            raise ValueError('%r is not a valid implementation. ' % (value,))
        self._tdl_noise_c.implementation = value

    @property
    def hurst(self):
        return self.noise_c.H

    @property
    def lacunarity(self):
        return self.noise_c.lacunarity

    @property
    def octaves(self):
        return self._tdl_noise_c.octaves
    @octaves.setter
    def octaves(self, value):
        self._tdl_noise_c.octaves = value

    def get_point(self, x=0, y=0, z=0, w=0):
        """Return the noise value at the (x, y, z, w) point.

        Args:
            x (float): The position on the 1st axis.
            y (float): The position on the 2nd axis.
            z (float): The position on the 3rd axis.
            w (float): The position on the 4th axis.
        """
        return lib.NoiseGetSample(self._tdl_noise_c, (x, y, z, w))

    def sample_mgrid(self, mgrid):
        """Sample a mesh-grid array and return the result.

        The :any:`sample_ogrid` method performs better as there is a lot of
        overhead when working with large mesh-grids.

        Args:
            mgrid (numpy.ndarray): A mesh-grid array of points to sample.
                A contiguous array of type `numpy.float32` is preferred.

        Returns:
            numpy.ndarray: An array of sampled points.

                This array has the shape: ``mgrid.shape[:-1]``.
                The ``dtype`` is `numpy.float32`.
        """
        mgrid = np.ascontiguousarray(mgrid, np.float32)
        if mgrid.shape[0] != self.dimensions:
            raise ValueError('mgrid.shape[0] must equal self.dimensions, '
                             '%r[0] != %r' % (mgrid.shape, self.dimensions))
        out = np.ndarray(mgrid.shape[1:], np.float32)
        if mgrid.shape[1:] != out.shape:
            raise ValueError('mgrid.shape[1:] must equal out.shape, '
                             '%r[1:] != %r' % (mgrid.shape, out.shape))
        lib.NoiseSampleMeshGrid(self._tdl_noise_c, out.size,
                                ffi.cast('float*', mgrid.ctypes.data),
                                ffi.cast('float*', out.ctypes.data))
        return out

    def sample_ogrid(self, ogrid):
        """Sample an open mesh-grid array and return the result.

        Args
            ogrid (Sequence[Sequence[float]]): An open mesh-grid.

        Returns:
            numpy.ndarray: An array of sampled points.

                The ``shape`` is based on the lengths of the open mesh-grid
                arrays.
                The ``dtype`` is `numpy.float32`.
        """
        if len(ogrid) != self.dimensions:
            raise ValueError('len(ogrid) must equal self.dimensions, '
                             '%r != %r' % (len(ogrid), self.dimensions))
        ogrids = [np.ascontiguousarray(array, np.float32) for array in ogrid]
        out = np.ndarray([array.size for array in ogrids], np.float32)
        lib.NoiseSampleOpenMeshGrid(
            self._tdl_noise_c,
            len(ogrids),
            out.shape,
            [ffi.cast('float*', array.ctypes.data) for array in ogrids],
            ffi.cast('float*', out.ctypes.data),
            )
        return out

    def __getstate__(self):
        state = self.__dict__.copy()
        if self.dimensions < 4 and self.noise_c.waveletTileData == ffi.NULL:
            # Trigger a side effect of wavelet, so that copies will be synced.
            saved_algo = self.algorithm
            self.algorithm = tcod.libtcod.NOISE_WAVELET
            self.get_point()
            self.algorithm = saved_algo

        waveletTileData = None
        if self.noise_c.waveletTileData != ffi.NULL:
            waveletTileData = list(self.noise_c.waveletTileData[0:32*32*32])
            state['_waveletTileData'] = waveletTileData

        state['noise_c'] = {
            'ndim': self.noise_c.ndim,
            'map': list(self.noise_c.map),
            'buffer': [list(sub_buffer) for sub_buffer in self.noise_c.buffer],
            'H': self.noise_c.H,
            'lacunarity': self.noise_c.lacunarity,
            'exponent': list(self.noise_c.exponent),
            'waveletTileData': waveletTileData,
            'noise_type': self.noise_c.noise_type,
            }
        state['_tdl_noise_c'] = {
            'dimensions': self._tdl_noise_c.dimensions,
            'implementation': self._tdl_noise_c.implementation,
            'octaves': self._tdl_noise_c.octaves,
            }
        return state

    def __setstate__(self, state):
        if isinstance(state, tuple): # deprecated format
            return self._setstate_old(state)
        # unpack wavelet tile data if it exists
        if '_waveletTileData' in state:
            state['_waveletTileData'] = ffi.new('float[]',
                                                state['_waveletTileData'])
            state['noise_c']['waveletTileData'] = state['_waveletTileData']
        else:
            state['noise_c']['waveletTileData'] = ffi.NULL

        # unpack perlin_data_t and link to Random instance
        state['noise_c']['rand'] = state['_random'].random_c
        state['noise_c'] = ffi.new('perlin_data_t*', state['noise_c'])

        # unpack TDLNoise and link to libtcod noise
        state['_tdl_noise_c']['noise'] = state['noise_c']
        state['_tdl_noise_c'] = ffi.new('TDLNoise*', state['_tdl_noise_c'])
        self.__dict__.update(state)

    def _setstate_old(self, state):
        self._random = state[0]
        self.noise_c = ffi.new('perlin_data_t*')
        self.noise_c.ndim = state[3]
        ffi.buffer(self.noise_c.map)[:] = state[4]
        ffi.buffer(self.noise_c.buffer)[:] = state[5]
        self.noise_c.H = state[6]
        self.noise_c.lacunarity = state[7]
        ffi.buffer(self.noise_c.exponent)[:] = state[8]
        if state[9]:
            # high change of this being prematurely garbage collected!
            self.__waveletTileData = ffi.new('float[]', 32*32*32)
            ffi.buffer(self.__waveletTileData)[:] = state[9]
        self.noise_c.noise_type = state[10]
        self._tdl_noise_c = ffi.new('TDLNoise*',
                                    (self.noise_c, self.noise_c.ndim,
                                     state[1], state[2]))



"""
    This module provides advanced noise generation.

    Noise is sometimes used for over-world generation, height-maps, and
    cloud/mist/smoke effects among other things.

    You can see examples of the available noise algorithms in the libtcod
    documentation `here
    <http://doryen.eptalys.net/data/libtcod/doc/1.5.1/html2/noise.html>`_.
"""


import random as _random

from tcod import ffi as _ffi
from tcod import lib as _lib

import tdl as _tdl
from . import style as _style

_MERSENNE_TWISTER = 1
_CARRY_WITH_MULTIPLY = 2

_MAX_DIMENSIONS = 4
_MAX_OCTAVES = 128

_NOISE_TYPES = {'PERLIN': 1, 'SIMPLEX': 2, 'WAVELET': 4}
_NOISE_MODES = {'FLAT': _lib.TCOD_noise_get,
                'FBM': _lib.TCOD_noise_get_fbm,
                'TURBULENCE': _lib.TCOD_noise_get_turbulence}

class Noise(object):
    """An advanced noise generator.

    .. deprecated:: 3.2
        This class has been replaced by :any:`tcod.noise.Noise`.

    Args:
        algorithm (Text): The primary noise algorithm to be used.

            Can be one of 'PERLIN', 'SIMPLEX', 'WAVELET'

            - 'PERLIN' -
              A popular noise generator.
            - 'SIMPLEX' -
              In theory this is a slightly faster generator with
              less noticeable directional artifacts.
            - 'WAVELET' -
              A noise generator designed to reduce aliasing and
              not lose detail when summed into a fractal
              (as with the 'FBM' and 'TURBULENCE' modes.)
              This works faster at higher dimensions.

        mode (Text): A secondary parameter to determine how noise is generated.

            Can be one of 'FLAT', 'FBM', 'TURBULENCE'

            - 'FLAT' -
              Generates the simplest form of noise.
              This mode does not use the hurst, lacunarity,
              and octaves parameters.
            - 'FBM' -
              Generates fractal brownian motion.
            - 'TURBULENCE' -
              Generates detailed noise with smoother and more
              natural transitions.

        hurst (float): The hurst exponent.

            This describes the raggedness of the resultant noise,
            with a higher value leading to a smoother noise.
            It should be in the 0.0-1.0 range.

            This is only used in 'FBM' and 'TURBULENCE' modes.

        lacunarity (float): A multiplier that determines how quickly the
            frequency increases for each successive octave.

            The frequency of each successive octave is equal to
            the product of the previous octave's frequency and
            the lacunarity value.

            This is only used in 'FBM' and 'TURBULENCE' modes.

        octaves (float): Controls the amount of detail in the noise.

            This is only used in 'FBM' and 'TURBULENCE' modes.

        seed (Hashable): You can use any hashable object to be a seed for the
            noise generator.

            If None is used then a random seed will be generated.
    """

    def __init__(self, algorithm='PERLIN', mode='FLAT',
                 hurst=0.5, lacunarity=2.0, octaves=4.0, seed=None, dimensions=4):
        if algorithm.upper() not in _NOISE_TYPES:
            raise _tdl.TDLError('No such noise algorithm as %s' % algorithm)
        self._algorithm = algorithm.upper()

        if mode.upper() not in _NOISE_MODES:
            raise _tdl.TDLError('No such mode as %s' % mode)
        self._mode = mode.upper()

        if seed is None:
            seed = _random.getrandbits(32)
        try:
            seed = int(seed)
        except TypeError:
            seed = hash(seed)
        self._seed = seed
        # convert values into ctypes to speed up later functions
        self._dimensions = min(_MAX_DIMENSIONS, int(dimensions))
        if self._algorithm == 'WAVELET':
            self._dimensions = min(self._dimensions, 3) # Wavelet only goes up to 3
        self._random = _lib.TCOD_random_new_from_seed(
            _MERSENNE_TWISTER,
            _ffi.cast('uint32_t', self._seed),
            )
        self._hurst = hurst
        self._lacunarity = lacunarity
        self._noise = _lib.TCOD_noise_new(self._dimensions, self._hurst,
                                          self._lacunarity, self._random)
        _lib.TCOD_noise_set_type(self._noise, _NOISE_TYPES[self._algorithm])
        self._noiseFunc = _NOISE_MODES[self._mode]
        self._octaves = octaves
        self._useOctaves = (self._mode != 'FLAT')
        self._arrayType = 'float[%i]' % self._dimensions
        #self._cFloatArray = _ctypes.c_float * self._dimensions
        #self._array = self._cFloatArray()

    def __copy__(self):
        # using the pickle method is a convenient way to clone this object
        return self.__class__(*self.__getstate__())

    def __getstate__(self):
        return (self._algorithm, self._mode,
                self._hurst, self._lacunarity, self._octaves,
                self._seed, self._dimensions)

    def __setstate__(self, state):
        self.__init__(*state)

    def get_point(self, *position):
        """Return the noise value of a specific position.

        Example usage: value = noise.getPoint(x, y, z)

        Args:
            position (Tuple[float, ...]): The point to sample at.

        Returns:
            float: The noise value at position.

                This will be a floating point in the 0.0-1.0 range.
        """
        #array = self._array
        #for d, pos in enumerate(position):
        #    array[d] = pos
        #array = self._cFloatArray(*position)
        array = _ffi.new(self._arrayType, position)
        if self._useOctaves:
            return (self._noiseFunc(self._noise, array, self._octaves) + 1) * 0.5
        return (self._noiseFunc(self._noise, array) + 1) * 0.5


__all__ = [_var for _var in locals().keys() if _var[0] != '_']

Noise.getPoint = _style.backport(Noise.get_point)



#!/usr/bin/env python3

import os
import sys

import glob

from cffi import FFI
from pycparser import c_parser, c_ast, parse_file, c_generator

import shutil
import subprocess
import platform
try:
    from urllib import urlretrieve
except ImportError:
    from urllib.request import urlretrieve
import zipfile

SDL2_VERSION = os.environ.get('SDL_VERSION', '2.0.8')
TDL_NO_SDL2_EXPORTS = os.environ.get('TDL_NO_SDL2_EXPORTS', '0') == '1'

CFFI_HEADER = 'tcod/cffi.h'
CFFI_EXTRA_CDEFS = 'tcod/cdef.h'

BITSIZE, LINKAGE = platform.architecture()

def walk_sources(directory):
    for path, dirs, files in os.walk(directory):
        for source in files:
            if not source.endswith('.c'):
                continue
            yield os.path.join(path, source)

def find_sources(directory):
    return [os.path.join(directory, source)
            for source in os.listdir(directory)
            if source.endswith('.c')]

def get_sdl2_file(version):
    if sys.platform == 'win32':
        sdl2_file = 'SDL2-devel-%s-VC.zip' % (version,)
    else:
        assert sys.platform == 'darwin'
        sdl2_file = 'SDL2-%s.dmg' % (version,)
    sdl2_local_file = os.path.join('dependencies', sdl2_file)
    sdl2_remote_file = 'https://www.libsdl.org/release/%s' % sdl2_file
    if not os.path.exists(sdl2_local_file):
        print('Downloading %s' % sdl2_remote_file)
        urlretrieve(sdl2_remote_file, sdl2_local_file)
    return sdl2_local_file

def unpack_sdl2(version):
    sdl2_path = 'dependencies/SDL2-%s' % (version,)
    if sys.platform == 'darwin':
        sdl2_dir = sdl2_path
        sdl2_path += '/SDL2.framework'
    if os.path.exists(sdl2_path):
        return sdl2_path
    sdl2_arc = get_sdl2_file(version)
    print('Extracting %s' % sdl2_arc)
    if sdl2_arc.endswith('.zip'):
        with zipfile.ZipFile(sdl2_arc) as zf:
            zf.extractall('dependencies/')
    else:
        assert sdl2_arc.endswith('.dmg')
        subprocess.check_call(['hdiutil', 'mount', sdl2_arc])
        subprocess.check_call(['mkdir', '-p', sdl2_dir])
        subprocess.check_call(['cp', '-r', '/Volumes/SDL2/SDL2.framework',
                                           sdl2_dir])
        subprocess.check_call(['hdiutil', 'unmount', '/Volumes/SDL2'])
    return sdl2_path

module_name = 'tcod._libtcod'
include_dirs = [
    '.',
    'libtcod/include/',
    'libtcod/src/png/',
    'libtcod/src/zlib/',
]

extra_parse_args = []
extra_compile_args = []
extra_link_args = []
sources = []

libraries = []
library_dirs = []
define_macros = []

sources += walk_sources('tcod/')
sources += walk_sources('tdl/')
sources += ['libtcod/src/libtcod_c.c']
sources += ['libtcod/src/png/lodepng.c']
sources += glob.glob('libtcod/src/zlib/*.c')

if TDL_NO_SDL2_EXPORTS:
    extra_parse_args.append('-DTDL_NO_SDL2_EXPORTS')

if sys.platform == 'win32':
    libraries += ['User32', 'OpenGL32']
    define_macros.append(('TCODLIB_API', ''))
    define_macros.append(('_CRT_SECURE_NO_WARNINGS', None))

if 'linux' in sys.platform:
    libraries += ['GL']

if sys.platform == 'darwin':
    extra_link_args += ['-framework', 'OpenGL']
    extra_link_args += ['-framework', 'SDL2']
else:
    libraries += ['SDL2']

# included SDL headers are for whatever OS's don't easily come with them

if sys.platform in ['win32', 'darwin']:
    SDL2_PATH = unpack_sdl2(SDL2_VERSION)
    include_dirs.append('libtcod/src/zlib/')

if sys.platform == 'win32':
    include_dirs.append(os.path.join(SDL2_PATH, 'include'))
    ARCH_MAPPING = {'32bit': 'x86', '64bit': 'x64'}
    SDL2_LIB_DIR = os.path.join(SDL2_PATH, 'lib/', ARCH_MAPPING[BITSIZE])
    library_dirs.append(SDL2_LIB_DIR)
    SDL2_LIB_DEST = os.path.join('tcod', ARCH_MAPPING[BITSIZE])
    if not os.path.exists(SDL2_LIB_DEST):
        os.mkdir(SDL2_LIB_DEST)
    shutil.copy(os.path.join(SDL2_LIB_DIR, 'SDL2.dll'), SDL2_LIB_DEST)

def fix_header(filepath):
    """Removes leading whitespace from a MacOS header file.

    This whitespace is causing issues with directives on some platforms.
    """
    with open(filepath, 'r+') as f:
        current = f.read()
        fixed = '\n'.join(line.strip() for line in current.split('\n'))
        if current == fixed:
            return
        f.seek(0)
        f.truncate()
        f.write(fixed)

if sys.platform == 'darwin':
    HEADER_DIR = os.path.join(SDL2_PATH, 'Headers')
    fix_header(os.path.join(HEADER_DIR, 'SDL_assert.h'))
    fix_header(os.path.join(HEADER_DIR, 'SDL_config_macosx.h'))
    include_dirs.append(HEADER_DIR)
    extra_link_args += ['-F%s/..' % SDL2_PATH]
    extra_link_args += ['-rpath', '%s/..' % SDL2_PATH]
    extra_link_args += ['-rpath', '/usr/local/opt/llvm/lib/']

if sys.platform not in ['win32', 'darwin']:
    extra_parse_args += subprocess.check_output(['sdl2-config', '--cflags'],
                                              universal_newlines=True
                                              ).strip().split()
    extra_compile_args += extra_parse_args
    extra_link_args += subprocess.check_output(['sdl2-config', '--libs'],
                                               universal_newlines=True
                                               ).strip().split()

class CustomPostParser(c_ast.NodeVisitor):

    def __init__(self):
        self.ast = None
        self.typedefs = None

    def parse(self, ast):
        self.ast = ast
        self.typedefs = []
        self.visit(ast)
        return ast

    def visit_Typedef(self, node):
        if node.name in ['wchar_t', 'size_t']:
            # remove fake typedef placeholders
            self.ast.ext.remove(node)
        else:
            self.generic_visit(node)
            if node.name in self.typedefs:
                print('warning: %s redefined' % node.name)
                self.ast.ext.remove(node)
            self.typedefs.append(node.name)

    def visit_EnumeratorList(self, node):
        """Replace enumerator expressions with '...' stubs."""
        for type, enum in node.children():
            if enum.value is None:
                pass
            elif isinstance(enum.value, (c_ast.BinaryOp, c_ast.UnaryOp)):
                enum.value = c_ast.Constant('int', '...')
            elif hasattr(enum.value, 'type'):
                enum.value = c_ast.Constant(enum.value.type, '...')

    def visit_ArrayDecl(self, node):
        if not node.dim:
            return
        if isinstance(node.dim, (c_ast.BinaryOp, c_ast.UnaryOp)):
            node.dim = c_ast.Constant('int', '...')

    def visit_Decl(self, node):
        if node.name is None:
            self.generic_visit(node)
        elif (node.name and 'vsprint' in node.name or
              node.name in ['SDL_vsscanf',
                            'SDL_vsnprintf',
                            'SDL_LogMessageV']):
            # exclude va_list related functions
            self.ast.ext.remove(node)
        elif node.name in ['screen']:
            # exclude outdated 'extern SDL_Surface* screen;' line
            self.ast.ext.remove(node)
        else:
            self.generic_visit(node)

    def visit_FuncDef(self, node):
        """Exclude function definitions.  Should be declarations only."""
        self.ast.ext.remove(node)

def get_cdef():
    generator = c_generator.CGenerator()
    return generator.visit(get_ast())

def get_ast():
    global extra_parse_args
    if 'win32' in sys.platform:
        extra_parse_args += [r'-I%s/include' % SDL2_PATH]
    if 'darwin' in sys.platform:
        extra_parse_args += [r'-I%s/Headers' % SDL2_PATH]

    ast = parse_file(filename=CFFI_HEADER, use_cpp=True,
                     cpp_args=[r'-Idependencies/fake_libc_include',
                               r'-Ilibtcod/include',
                               r'-DDECLSPEC=',
                               r'-DSDLCALL=',
                               r'-DTCODLIB_API=',
                               r'-DSDL_FORCE_INLINE=',
                               r'-U__GNUC__',
                               r'-D_SDL_thread_h',
                               r'-DDOXYGEN_SHOULD_IGNORE_THIS',
                               r'-DMAC_OS_X_VERSION_MIN_REQUIRED=1060',
                               r'-D__attribute__(x)=',
                               r'-D_PSTDINT_H_INCLUDED',
                               ] + extra_parse_args)
    ast = CustomPostParser().parse(ast)
    return ast

# Can force the use of OpenMP with this variable.
try:
    USE_OPENMP = eval(os.environ.get('USE_OPENMP', 'None').title())
except Exception:
    USE_OPENMP = None

tdl_build = os.environ.get('TDL_BUILD', 'RELEASE').upper()

MSVC_CFLAGS = {
    'DEBUG': ['/Od'],
    'RELEASE': ['/GL', '/O2', '/GS-'],
}
MSVC_LDFLAGS = {
    'DEBUG': [],
    'RELEASE': ['/LTCG'],
}
GCC_CFLAGS = {
    'DEBUG': ['-O0'],
    'RELEASE': ['-flto', '-O3', '-fPIC'],
}

if sys.platform == 'win32' and '--compiler=mingw32' not in sys.argv:
    extra_compile_args.extend(MSVC_CFLAGS[tdl_build])
    extra_link_args.extend(MSVC_LDFLAGS[tdl_build])

    if USE_OPENMP is None:
        USE_OPENMP = sys.version_info[:2] >= (3, 5)

    if USE_OPENMP:
        extra_compile_args.append('/openmp')
else:
    extra_compile_args.extend(GCC_CFLAGS[tdl_build])
    extra_link_args.extend(GCC_CFLAGS[tdl_build])
    if USE_OPENMP is None:
        USE_OPENMP = sys.platform != 'darwin'

    if USE_OPENMP:
        extra_compile_args.append('-fopenmp')
        extra_link_args.append('-fopenmp')

ffi = FFI()
ffi.cdef(get_cdef())
ffi.cdef(open(CFFI_EXTRA_CDEFS, 'r').read())
ffi.set_source(
    module_name, '#include <tcod/cffi.h>',
    include_dirs=include_dirs,
    library_dirs=library_dirs,
    sources=sources,
    libraries=libraries,
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    define_macros=define_macros,
)

if __name__ == "__main__":
    ffi.compile()



"""

"""

from __future__ import absolute_import

from tcod.libtcod import ffi, lib


class Color(list):
    """

    Args:
        r (int): Red value, from 0 to 255.
        g (int): Green value, from 0 to 255.
        b (int): Blue value, from 0 to 255.
    """

    def __init__(self, r=0, g=0, b=0):
        self[:] = (r & 0xff, g & 0xff, b & 0xff)

    @property
    def r(self):
        """int: Red value, always normalised to 0-255."""
        return self[0]

    @r.setter
    def r(self, value):
        self[0] = value & 0xff

    @property
    def g(self):
        """int: Green value, always normalised to 0-255."""
        return self[1]
    @g.setter
    def g(self, value):
        self[1] = value & 0xff

    @property
    def b(self):
        """int: Blue value, always normalised to 0-255."""
        return self[2]
    @b.setter
    def b(self, value):
        self[2] = value & 0xff

    @classmethod
    def _new_from_cdata(cls, cdata):
        """new in libtcod-cffi"""
        return cls(cdata.r, cdata.g, cdata.b)

    def __getitem__(self, index):
        try:
            return list.__getitem__(self, index)
        except TypeError:
            return list.__getitem__(self, 'rgb'.index(index))

    def __setitem__(self, index, value):
        try:
            list.__setitem__(self, index, value)
        except TypeError:
            list.__setitem__(self, 'rgb'.index(index), value)

    def __eq__(self, other):
        """Compare equality between colors.

        Also compares with standard sequences such as 3-item tuples or lists.
        """
        try:
            return bool(lib.TCOD_color_equals(self, other))
        except TypeError:
            return False

    def __add__(self, other):
        """Add two colors together."""
        return Color._new_from_cdata(lib.TCOD_color_add(self, other))

    def __sub__(self, other):
        """Subtract one color from another."""
        return Color._new_from_cdata(lib.TCOD_color_subtract(self, other))

    def __mul__(self, other):
        """Multiply with a scaler or another color."""
        if isinstance(other, (Color, list, tuple)):
            return Color._new_from_cdata(lib.TCOD_color_multiply(self, other))
        else:
            return Color._new_from_cdata(
                lib.TCOD_color_multiply_scalar(self, other))

    def __repr__(self):
        """Return a printable representation of the current color."""
        return "%s(%i,%i,%i)" % (self.__class__.__name__,
                                 self.r, self.g, self.b)


def _import_colors(lib):
    """Import all Color constants from lib into this module."""
    g = globals()
    for name in dir(lib):
        if name[:5] != 'TCOD_':
            continue
        value = getattr(lib, name)
        if not isinstance(value, ffi.CData):
            continue
        if ffi.typeof(value) != ffi.typeof('TCOD_color_t'):
            continue
        g[name[5:]] = Color._new_from_cdata(value)


_import_colors(lib)



"""libtcod map attributes and field-of-view functions.

Example::

    >>> import tcod.map
    >>> m = tcod.map.Map(width=3, height=4)
    >>> m.walkable
    array([[False, False, False],
           [False, False, False],
           [False, False, False],
           [False, False, False]]...)

    # Like the rest of the tcod modules, all arrays here are
    # in row-major order and are addressed with [y,x]
    >>> m.transparent[:] = True # Sets all to True.
    >>> m.transparent[1:3,0] = False # Sets (1, 0) and (2, 0) to False.
    >>> m.transparent
    array([[ True,  True,  True],
           [False,  True,  True],
           [False,  True,  True],
           [ True,  True,  True]]...)

    >>> m.compute_fov(0, 0)
    >>> m.fov
    array([[ True,  True,  True],
           [ True,  True,  True],
           [False,  True,  True],
           [False, False,  True]]...)
    >>> m.fov[3,1]
    False

"""

from __future__ import absolute_import

import numpy as np

from tcod.libtcod import lib, ffi
import tcod._internal


class Map(object):
    """A map containing libtcod attributes.

    .. versionchanged:: 4.1
        `transparent`, `walkable`, and `fov` are now numpy boolean arrays.

    .. versionchanged:: 4.3
        Added `order` parameter.

    Args:
        width (int): Width of the new Map.
        height (int): Height of the new Map.
        order (str): Which numpy memory order to use.

    Attributes:
        width (int): Read only width of this Map.
        height (int): Read only height of this Map.
        transparent: A boolean array of transparent cells.
        walkable: A boolean array of walkable cells.
        fov: A boolean array of the cells lit by :any:'compute_fov'.

    """

    def __init__(self, width, height, order='C'):
        assert ffi.sizeof('cell_t') == 3 # assert buffer alignment
        self.width = width
        self.height = height
        self._order = tcod._internal.verify_order(order)

        self.__buffer = np.zeros((height, width, 3), dtype=np.bool_)
        self.map_c = self.__as_cdata()


    def __as_cdata(self):
        return ffi.new(
            'map_t *',
            (
                self.width,
                self.height,
                self.width * self.height,
                ffi.cast('cell_t*', self.__buffer.ctypes.data),
            )
        )

    @property
    def transparent(self):
        buffer = self.__buffer[:,:,0]
        return buffer.T if self._order == 'F' else buffer

    @property
    def walkable(self):
        buffer = self.__buffer[:,:,1]
        return buffer.T if self._order == 'F' else buffer

    @property
    def fov(self):
        buffer = self.__buffer[:,:,2]
        return buffer.T if self._order == 'F' else buffer

    def compute_fov(self, x, y, radius=0, light_walls=True,
                    algorithm=lib.FOV_RESTRICTIVE):
        """Compute a field-of-view on the current instance.

        Args:
            x (int): Point of view, x-coordinate.
            y (int): Point of view, y-coordinate.
            radius (int): Maximum view distance from the point of view.

                A value of `0` will give an infinite distance.
            light_walls (bool): Light up walls, or only the floor.
            algorithm (int): Defaults to tcod.FOV_RESTRICTIVE
        """
        lib.TCOD_map_compute_fov(
            self.map_c, x, y, radius, light_walls, algorithm)

    def __setstate__(self, state):
        if '_Map__buffer' not in state: # deprecated
            # remove this check on major version update
            self.__buffer = np.zeros((state['height'], state['width'], 3),
                              dtype=np.bool_)
            self.__buffer[:,:,0] = state['buffer'] & 0x01
            self.__buffer[:,:,1] = state['buffer'] & 0x02
            self.__buffer[:,:,2] = state['buffer'] & 0x04
            del state['buffer']
            state['_order'] = 'F'
        if '_order' not in state: # remove this check on major version update
            raise RuntimeError(
                'This Map was saved with a bad version of tdl.')
        self.__dict__.update(state)
        self.map_c = self.__as_cdata()

    def __getstate__(self):
        state = self.__dict__.copy()
        del state['map_c']
        return state



#!/usr/bin/env python

import sys

from setuptools import setup

from subprocess import check_output
import platform

def get_version():
    """Get the current version from a git tag, or by reading tdl/version.py"""
    try:
        tag = check_output(['git', 'describe', '--abbrev=0'],
                           universal_newlines=True).strip()
        assert tag.startswith('v')
        version = tag[1:] # remove `v`

        # add .devNN if needed
        log = check_output(['git', 'log', '%s..HEAD' % tag, '--oneline'],
                           universal_newlines=True)
        commits_since_tag = log.count('\n')
        if commits_since_tag:
            version += '.dev%i' % commits_since_tag

        # update tdl/version.py
        open('tdl/version.py', 'w').write('__version__ = %r\n' % version)
        return version
    except:
        exec(open('tdl/version.py').read(), globals())
        return __version__

is_pypy = platform.python_implementation() == 'PyPy'

def get_package_data():
    '''get data files which will be included in the main tcod/ directory'''
    BITSIZE, LINKAGE = platform.architecture()
    files = [
        'lib/LIBTCOD-CREDITS.txt',
        'lib/LIBTCOD-LICENSE.txt',
        'lib/README-SDL.txt'
        ]
    if 'win32' in sys.platform:
        if BITSIZE == '32bit':
            files += ['x86/SDL2.dll']
        else:
            files += ['x64/SDL2.dll']
    if sys.platform == 'darwin':
        files += ['SDL2.framework/Versions/A/SDL2']
    return files

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

setup(
    name='tdl',
    version=get_version(),
    author='Kyle Stewart',
    author_email='4B796C65+tdl@gmail.com',
    description='Pythonic cffi port of libtcod.',
    long_description='\n'.join([open('README.rst', 'r').read(),
                                open('CHANGELOG.rst', 'r').read()]),
    url='https://github.com/HexDecimal/python-tdl',
    download_url='https://pypi.python.org/pypi/tdl',
    py_modules=['libtcodpy'],
    packages=['tdl', 'tcod'],
    package_data={
        'tdl': ['*.png'],
        'tcod': get_package_data(),
        },
    install_requires=[
        'cffi>=1.8.1,<2',
        'numpy>=1.10,<2' if not is_pypy else '',
        ],
    cffi_modules=['build_libtcod.py:ffi'],
    setup_requires=[
        'cffi>=1.8.1,<2',
        'pycparser>=2.14,<3',
        ] + pytest_runner,
    tests_require=[
        'pytest',
        'pytest-cov',
        'pytest-benchmark',
        ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Win32 (MS Windows)',
        'Environment :: MacOS X',
        'Environment :: X11 Applications',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Games/Entertainment',
        'Topic :: Multimedia :: Graphics',
        'Topic :: Software Development :: Libraries :: Python Modules',
        ],
    keywords='roguelike cffi Unicode libtcod fov heightmap namegen',
    platforms=[
        'Windows',
        'MacOS',
        'Linux',
        ],
    license='Simplified BSD License',
    )
